# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/tf_object_detection/keypoint_box_coder.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/tf_object_detection/keypoint_box_coder.proto',
  package='rastervision.protos.tf_object_detection',
  syntax='proto2',
  serialized_pb=_b('\n@rastervision/protos/tf_object_detection/keypoint_box_coder.proto\x12\'rastervision.protos.tf_object_detection\"\x84\x01\n\x10KeypointBoxCoder\x12\x15\n\rnum_keypoints\x18\x01 \x01(\x05\x12\x13\n\x07y_scale\x18\x02 \x01(\x02:\x02\x31\x30\x12\x13\n\x07x_scale\x18\x03 \x01(\x02:\x02\x31\x30\x12\x17\n\x0cheight_scale\x18\x04 \x01(\x02:\x01\x35\x12\x16\n\x0bwidth_scale\x18\x05 \x01(\x02:\x01\x35')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_KEYPOINTBOXCODER = _descriptor.Descriptor(
  name='KeypointBoxCoder',
  full_name='rastervision.protos.tf_object_detection.KeypointBoxCoder',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_keypoints', full_name='rastervision.protos.tf_object_detection.KeypointBoxCoder.num_keypoints', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='y_scale', full_name='rastervision.protos.tf_object_detection.KeypointBoxCoder.y_scale', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(10),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='x_scale', full_name='rastervision.protos.tf_object_detection.KeypointBoxCoder.x_scale', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(10),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='height_scale', full_name='rastervision.protos.tf_object_detection.KeypointBoxCoder.height_scale', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(5),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='width_scale', full_name='rastervision.protos.tf_object_detection.KeypointBoxCoder.width_scale', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(5),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=110,
  serialized_end=242,
)

DESCRIPTOR.message_types_by_name['KeypointBoxCoder'] = _KEYPOINTBOXCODER

KeypointBoxCoder = _reflection.GeneratedProtocolMessageType('KeypointBoxCoder', (_message.Message,), dict(
  DESCRIPTOR = _KEYPOINTBOXCODER,
  __module__ = 'rastervision.protos.tf_object_detection.keypoint_box_coder_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.KeypointBoxCoder)
  ))
_sym_db.RegisterMessage(KeypointBoxCoder)


# @@protoc_insertion_point(module_scope)



from abc import abstractmethod
from copy import deepcopy

import rastervision as rv
from rastervision.rv_config import RVConfig
from rastervision.core import (Config, ConfigBuilder, BundledConfigMixin)


class BackendConfig(BundledConfigMixin, Config):
    def __init__(self, backend_type, pretrained_model_uri=None):
        self.backend_type = backend_type
        self.pretrained_model_uri = pretrained_model_uri

    @abstractmethod
    def create_backend(self, task_config):
        """Create the Backend that this configuration represents

           Args:
              task_config: The task configuration for the task
                           to be accomplished by this backend.
        """
        pass

    def to_builder(self):
        return rv._registry.get_config_builder(rv.BACKEND,
                                               self.backend_type)(self)

    @staticmethod
    def builder(backend_type):
        return rv._registry.get_config_builder(rv.BACKEND, backend_type)()

    @staticmethod
    def from_proto(msg):
        """Creates a BackendConfig from the specificed protobuf message
        """
        return rv._registry.get_config_builder(rv.BACKEND, msg.backend_type)() \
                           .from_proto(msg) \
                           .build()

    def report_io(self, command_type, io_def):
        if command_type == rv.TRAIN:
            if self.pretrained_model_uri:
                io_def.add_input(self.pretrained_model_uri)


class BackendConfigBuilder(ConfigBuilder):
    def __init__(self, backend_type, config_class, config=None, prev=None):
        if config is None:
            config = {}
        super().__init__(config_class, config)
        self.task = None
        self.backend_type = backend_type

    @abstractmethod
    def _applicable_tasks(self):
        """Returns the tasks that this backend can be applied to.
        """
        pass

    @abstractmethod
    def _process_task(self, task):
        """Subclasses override this to set up configuration related
           to this task
        """
        pass

    def from_proto(self, msg):
        return self \
            .with_pretrained_model(msg.pretrained_model_uri)

    def with_task(self, task):
        """Sets a specific task type.

        Args:
            task:  A TaskConfig object.

        """
        if task.task_type not in self._applicable_tasks():
            raise Exception(
                'Backend of type {} cannot be applied to task type {}'.format(
                    self.backend_type, task.task_type))
        b = deepcopy(self)
        b.task = task
        b = b._process_task()
        return b

    def with_pretrained_model(self, uri):
        """Set a pretrained model URI. The filetype and meaning
           for this model will be different based on the backend implementation.
        """
        b = deepcopy(self)
        b.config['pretrained_model_uri'] = uri
        return b

    def with_model_defaults(self, model_defaults_key):
        """Sets the backend configuration and pretrained model defaults
           according to the model defaults configuration.
        """
        model_defaults = RVConfig.get_instance().get_model_defaults()

        if self.backend_type in model_defaults:
            backend_defaults = model_defaults[self.backend_type]
            if model_defaults_key in backend_defaults:
                return self._load_model_defaults(
                    backend_defaults[model_defaults_key])
            else:
                raise rv.ConfigError('No defaults found for model key {}'
                                     .format(model_defaults_key))
        else:
            raise rv.ConfigError('No model defaults for backend {}'
                                 .format(self.backend_type))
        return self

    def _load_model_defaults(self, model_defaults):
        """Overriding classes should handle this if they
           want to allow default parameters to be loaded
           from the default configurations.
        """
        return self



from typing import List
import logging

import numpy as np

from .task import Task
from rastervision.core.box import Box
from rastervision.data.scene import Scene
from rastervision.data.label import SemanticSegmentationLabels
from rastervision.core.training_data import TrainingData

TRAIN = 'train'
VALIDATION = 'validation'

log = logging.getLogger(__name__)


def get_random_sample_train_windows(label_store, chip_size, class_map, extent,
                                    chip_options, filter_windows):
    prob = chip_options.negative_survival_probability
    target_count_threshold = chip_options.target_count_threshold
    target_classes = chip_options.target_classes
    chips_per_scene = chip_options.chips_per_scene

    if not target_classes:
        all_class_ids = [item.id for item in class_map.get_items()]
        target_classes = all_class_ids

    windows = []
    attempts = 0
    while (attempts < chips_per_scene):
        candidate_window = extent.make_random_square(chip_size)
        if not filter_windows([candidate_window]):
            continue
        attempts = attempts + 1

        if (prob >= 1.0):
            windows.append(candidate_window)
        elif attempts == chips_per_scene and len(windows) == 0:
            windows.append(candidate_window)
        else:
            good = label_store.enough_target_pixels(
                candidate_window, target_count_threshold, target_classes)
            if good or (np.random.rand() < prob):
                windows.append(candidate_window)

    return windows


class SemanticSegmentation(Task):
    """Task-derived type that implements the semantic segmentation task."""

    def get_train_windows(self, scene: Scene) -> List[Box]:
        """Get training windows covering a scene.

        Args:
             scene: The scene over-which windows are to be generated.

        Returns:
             A list of windows, list(Box)

        """
        raster_source = scene.raster_source
        extent = raster_source.get_extent()
        label_source = scene.ground_truth_label_source
        chip_size = self.config.chip_size
        chip_options = self.config.chip_options

        def filter_windows(windows):
            if scene.aoi_polygons:
                windows = Box.filter_by_aoi(windows, scene.aoi_polygons)

            if 0 in self.config.class_map.get_keys():
                filt_windows = []
                for w in windows:
                    label_arr = label_source.get_labels(w).get_label_arr(w)
                    ignore_inds = label_arr.ravel() == 0
                    if np.all(ignore_inds):
                        pass
                    else:
                        filt_windows.append(w)
                windows = filt_windows
            return windows

        if chip_options.window_method == 'random_sample':
            return get_random_sample_train_windows(
                label_source, chip_size, self.config.class_map, extent,
                chip_options, filter_windows)
        elif chip_options.window_method == 'sliding':
            stride = chip_options.stride
            if stride is None:
                stride = chip_size / 2
            stride = int(round(stride))

            return list(
                filter_windows((extent.get_windows(chip_size, stride))))

    def make_chips(self, train_scenes, validation_scenes, augmentors, tmp_dir):
        """Make training chips.

        Convert Scenes with a ground_truth_label_store into training
        chips in MLBackend-specific format, and write to URI specified in
        options.

        Args:
            train_scenes: list of Scenes
            validation_scenes: list of Scenes
                (that is disjoint from train_scenes)
            augmentors: Augmentors used to augment training data
        """

        def _process_scene(scene, type_, augment):
            with scene.activate():
                data = TrainingData()
                log.info('Making {} chips for scene: {}'.format(
                    type_, scene.id))
                windows = self.get_train_windows(scene)
                for window in windows:
                    chip = scene.raster_source.get_chip(window)
                    labels = self.get_train_labels(window, scene)

                    # If chip has ignore labels, fill in those pixels with
                    # nodata.
                    label_arr = labels.get_label_arr(window)
                    zero_inds = label_arr.ravel() == 0
                    chip_shape = chip.shape
                    if np.any(zero_inds):
                        chip = np.reshape(chip, (-1, chip.shape[2]))
                        chip[zero_inds, :] = 0
                        chip = np.reshape(chip, chip_shape)

                    data.append(chip, window, labels)
                # Shuffle data so the first N samples which are displayed in
                # Tensorboard are more diverse.
                data.shuffle()

                # Process augmentation
                if augment:
                    for augmentor in augmentors:
                        data = augmentor.process(data, tmp_dir)

                return self.backend.process_scene_data(scene, data, tmp_dir)

        def _process_scenes(scenes, type_, augment):
            return [_process_scene(scene, type_, augment) for scene in scenes]

        processed_training_results = _process_scenes(
            train_scenes, TRAIN, augment=True)
        processed_validation_results = _process_scenes(
            validation_scenes, VALIDATION, augment=False)

        self.backend.process_sceneset_results(
            processed_training_results, processed_validation_results, tmp_dir)

    def get_train_labels(self, window: Box, scene: Scene) -> np.ndarray:
        """Get the training labels for the given window in the given scene.

        Args:
             window: The window over-which the labels are to be
                  retrieved.
             scene: The scene from-which the window of labels is to be
                  extracted.

        Returns:
             An appropriately-shaped 2d np.ndarray with the labels
             encoded as packed pixels.

        """
        label_store = scene.ground_truth_label_source
        return label_store.get_labels(window)

    def get_predict_windows(self, extent: Box) -> List[Box]:
        """Get windows over-which predictions will be calculated.

        Args:
             extent: The overall extent of the area.

        Returns:
             An sequence of windows.

        """
        chip_size = self.config.predict_chip_size
        return extent.get_windows(chip_size, chip_size)

    def post_process_predictions(self, labels, scene):
        return labels

    def save_debug_predict_image(self, scene, debug_dir_uri):
        # TODO implement this
        pass

    def predict_scene(self, scene, tmp_dir):
        """Predict on a single scene, and return the labels."""
        log.info('Making predictions for scene')
        raster_source = scene.raster_source
        windows = self.get_predict_windows(raster_source.get_extent())

        def label_fn(window):
            chip = raster_source.get_chip(window)
            labels = self.backend.predict([chip], [window], tmp_dir)
            label_arr = labels.get_label_arr(window)

            # Set NODATA pixels in imagery to predicted value of 0 (ie. ignore)
            label_arr[np.sum(chip, axis=2) == 0] = 0

            print('.', end='', flush=True)
            return label_arr

        return SemanticSegmentationLabels(windows, label_fn)



from rastervision.data.label import Labels

import numpy as np
from rasterio.features import rasterize
import shapely


class SemanticSegmentationLabels(Labels):
    """A set of spatially referenced semantic segmentation labels.

    Since labels are represented as rasters, the labels for a scene can take up a lot of
    memory. Therefore, to avoid running out of memory, labels are computed as needed for
    windows.
    """

    def __init__(self, windows, label_fn, aoi_polygons=None):
        """Constructor

        Args:
            windows: a list of Box representing the windows covering a scene
            label_fn: a function that takes a window (Box) and returns a label array
                of the same shape with each value a class id.
            aoi_polygons: a list of shapely.geom that contains the AOIs
                (areas of interest) for a scene.

        """
        self.windows = windows
        self.label_fn = label_fn
        self.aoi_polygons = aoi_polygons

    def __add__(self, other):
        """Add labels to these labels.

        Returns a concatenation of this and the other labels.
        """
        return SemanticSegmentationLabels(
            self.windows + other.windows,
            self.label_fn,
            aoi_polygons=self.aoi_polygons)

    def __eq__(self, other):
        for window in self.get_windows():
            if not np.array_equal(
                    self.get_label_arr(window), other.get_label_arr(window)):
                return False
        return True

    def filter_by_aoi(self, aoi_polygons):
        """Returns a new SemanticSegmentationLabels object with aoi_polygons set."""
        return SemanticSegmentationLabels(
            self.windows, self.label_fn, aoi_polygons=aoi_polygons)

    def add_window(self, window):
        self.windows.append(window)

    def get_windows(self):
        return self.windows

    def get_label_arr(self, window, clip_extent=None):
        """Get the label array for a window.

        Note: the window should be kept relatively small to avoid running out of memory.

        Args:
            window: Box
            clip_extent: a Box representing the extent of the corresponding Scene

        Returns:
            np.ndarray of class_ids with zeros filled in outside the AOIs and clipped
                to the clip_extent
        """
        window_geom = window.to_shapely()

        if not self.aoi_polygons:
            label_arr = self.label_fn(window)
        else:
            # For each aoi_polygon, intersect with window, and put in window frame of
            # reference.
            window_aois = []
            for aoi in self.aoi_polygons:
                window_aoi = aoi.intersection(window_geom)
                if not window_aoi.is_empty:

                    def transform_shape(x, y, z=None):
                        return (x - window.xmin, y - window.ymin)

                    window_aoi = shapely.ops.transform(transform_shape,
                                                       window_aoi)
                    window_aois.append(window_aoi)

            if window_aois:
                # If window intersects with AOI, set pixels outside the AOI polygon to 0,
                # so they are ignored during eval.
                label_arr = self.label_fn(window)
                mask = rasterize(
                    [(p, 0) for p in window_aois],
                    out_shape=label_arr.shape,
                    fill=1,
                    dtype=np.uint8)
                label_arr[mask.astype(np.bool)] = 0
            else:
                # If window does't overlap with any AOI, then return all zeros.
                label_arr = np.zeros((window.get_height(), window.get_width()))

        if clip_extent is not None:
            clip_window = window.intersection(clip_extent)
            label_arr = label_arr[0:clip_window.get_height(), 0:
                                  clip_window.get_width()]

        return label_arr



import rastervision as rv


class Verbosity:
    QUIET = 0
    NORMAL = 1
    VERBOSE = 2
    VERY_VERBOSE = 3
    DEBUG = 4

    @staticmethod
    def get():
        return rv._registry._get_rv_config().get_verbosity()



from abc import ABC, abstractmethod
from copy import deepcopy

import rastervision as rv
from rastervision.plugin import PluginRegistry
from rastervision.protos.command_pb2 \
    import CommandConfig as CommandConfigMsg


class CommandConfig(ABC):
    def __init__(self, command_type, root_uri, split_id=0):
        self.command_type = command_type
        self.root_uri = root_uri
        self.split_id = split_id

    @abstractmethod
    def create_command(self):
        """Run the command."""
        pass

    @abstractmethod
    def report_io(self):
        """Returns an IODefinition object with all inputs and outputs reported
        for this command.
        """
        pass

    def split(self, num_parts):
        """Split this command config into num_parts parts if possible.
        This will return a list of size between 1 and num_parts containing command
        configurations. If the command configuration does not split, will return [self]
        """
        return [self]

    def utilizes_gpu(self):
        """Method that determines if this command can utilize a GPU.
        This is useful for runners to know what resources to utilize
        for command execution - e.g. don't spin up the more expensive
        GPU machines if this command can't use the GPUs.

        Defaults to False.
        """
        return False

    def to_proto(self):
        """Returns the protobuf configuration for this config.
        """
        plugin_config = PluginRegistry.get_instance().to_proto()
        return CommandConfigMsg(
            command_type=self.command_type,
            root_uri=self.root_uri,
            split_id=self.split_id,
            plugins=plugin_config)

    def to_builder(self):
        return rv._registry.get_command_config_builder(self.command_type)(self)

    @staticmethod
    def builder(command_type):
        """Returns a new builder for the given command type.
        """
        return rv._registry.get_command_config_builder(command_type)()

    @staticmethod
    def from_proto(msg):
        """Creates a TaskConfig from the specificed protobuf message
        """
        command_type = msg.command_type
        return rv._registry.get_command_config_builder(command_type)() \
                           .from_proto(msg) \
                           .build()


class CommandConfigBuilder(ABC):
    def __init__(self, command_type, prev):
        self.command_type = command_type

        if prev is None:
            self.root_uri = None
            self.split_id = 0
        else:
            self.root_uri = prev.root_uri
            self.split_id = prev.split_id

    @abstractmethod
    def build(self, prev=None):
        """Returns the configuration that is built by this builder.
        """
        pass

    @abstractmethod
    def get_root_uri(self, experiment_config):
        """Return the root URI for this command for a given experiment"""
        pass

    def from_proto(self, msg):
        """Return a builder that takes the configuration from the proto message
           as its starting point.
        """

        # Process plugins from a command config protobuf message.
        if msg.HasField('plugins'):
            PluginRegistry.get_instance().add_plugins_from_proto(msg.plugins)

        return self.with_root_uri(msg.root_uri).with_split_id(msg.split_id)

    def validate(self):
        if self.root_uri is None:
            raise rv.ConfigError(
                'root_uri not set. Use with_root_uri or with_experiment')

    def with_experiment(self, experiment_config):
        """Generate all required information from this experiment.
           It is sufficient to only call 'with_experiment' before
           calling .build()
        """
        return self.with_root_uri(self.get_root_uri(experiment_config))

    def with_root_uri(self, root_uri):
        b = deepcopy(self)
        b.root_uri = root_uri
        return b

    def with_split_id(self, split_id):
        b = deepcopy(self)
        b.split_id = split_id
        return b



import numpy as np
from sklearn import metrics

from rastervision.evaluation import ClassificationEvaluation
from rastervision.evaluation import ClassEvaluationItem


class ChipClassificationEvaluation(ClassificationEvaluation):
    def __init__(self, class_map):
        super().__init__()
        self.class_map = class_map

    def compute(self, ground_truth_labels, prediction_labels):
        self.class_to_eval_item = ChipClassificationEvaluation.compute_eval_items(
            ground_truth_labels, prediction_labels, self.class_map)

        self.compute_avg()

    @staticmethod
    def compute_eval_items(gt_labels, pred_labels, class_map):
        nb_classes = len(class_map)
        class_to_eval_item = {}

        gt_class_ids = []
        pred_class_ids = []

        gt_cells = gt_labels.get_cells()
        for gt_cell in gt_cells:
            gt_class_id = gt_labels.get_cell_class_id(gt_cell)
            pred_class_id = pred_labels.get_cell_class_id(gt_cell)

            if gt_class_id is not None and pred_class_id is not None:
                gt_class_ids.append(gt_class_id)
                pred_class_ids.append(pred_class_id)

        # Add 1 because class_ids start at 1.
        sklabels = np.arange(1 + nb_classes)
        precision, recall, f1, support = metrics.precision_recall_fscore_support(
            gt_class_ids, pred_class_ids, labels=sklabels, warn_for=())

        for class_map_item in class_map.get_items():
            class_id = class_map_item.id
            class_name = class_map_item.name

            eval_item = ClassEvaluationItem(
                float(precision[class_id]),
                float(recall[class_id]),
                float(f1[class_id]),
                gt_count=float(support[class_id]),
                class_id=class_id,
                class_name=class_name)
            class_to_eval_item[class_id] = eval_item

        return class_to_eval_item



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/plugin.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/plugin.proto',
  package='rv.protos',
  syntax='proto2',
  serialized_pb=_b('\n rastervision/protos/plugin.proto\x12\trv.protos\";\n\x0cPluginConfig\x12\x13\n\x0bplugin_uris\x18\x01 \x03(\t\x12\x16\n\x0eplugin_modules\x18\x02 \x03(\t')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_PLUGINCONFIG = _descriptor.Descriptor(
  name='PluginConfig',
  full_name='rv.protos.PluginConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='plugin_uris', full_name='rv.protos.PluginConfig.plugin_uris', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='plugin_modules', full_name='rv.protos.PluginConfig.plugin_modules', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=47,
  serialized_end=106,
)

DESCRIPTOR.message_types_by_name['PluginConfig'] = _PLUGINCONFIG

PluginConfig = _reflection.GeneratedProtocolMessageType('PluginConfig', (_message.Message,), dict(
  DESCRIPTOR = _PLUGINCONFIG,
  __module__ = 'rastervision.protos.plugin_pb2'
  # @@protoc_insertion_point(class_scope:rv.protos.PluginConfig)
  ))
_sym_db.RegisterMessage(PluginConfig)


# @@protoc_insertion_point(module_scope)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/evaluator.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from rastervision.protos import class_item_pb2 as rastervision_dot_protos_dot_class__item__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/evaluator.proto',
  package='rv.protos',
  syntax='proto2',
  serialized_pb=_b('\n#rastervision/protos/evaluator.proto\x12\trv.protos\x1a$rastervision/protos/class_item.proto\x1a\x1cgoogle/protobuf/struct.proto\"\xc5\x02\n\x0f\x45valuatorConfig\x12\x16\n\x0e\x65valuator_type\x18\x01 \x02(\t\x12Y\n\x15\x63lassification_config\x18\x02 \x01(\x0b\x32\x38.rv.protos.EvaluatorConfig.ClassificationEvaluatorConfigH\x00\x12\x30\n\rcustom_config\x18\x03 \x01(\x0b\x32\x17.google.protobuf.StructH\x00\x1ay\n\x1d\x43lassificationEvaluatorConfig\x12\x12\n\noutput_uri\x18\x01 \x02(\t\x12\x19\n\x11vector_output_uri\x18\x03 \x01(\t\x12)\n\x0b\x63lass_items\x18\x02 \x03(\x0b\x32\x14.rv.protos.ClassItemB\x12\n\x10\x65valuator_config')
  ,
  dependencies=[rastervision_dot_protos_dot_class__item__pb2.DESCRIPTOR,google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_EVALUATORCONFIG_CLASSIFICATIONEVALUATORCONFIG = _descriptor.Descriptor(
  name='ClassificationEvaluatorConfig',
  full_name='rv.protos.EvaluatorConfig.ClassificationEvaluatorConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='output_uri', full_name='rv.protos.EvaluatorConfig.ClassificationEvaluatorConfig.output_uri', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='vector_output_uri', full_name='rv.protos.EvaluatorConfig.ClassificationEvaluatorConfig.vector_output_uri', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='class_items', full_name='rv.protos.EvaluatorConfig.ClassificationEvaluatorConfig.class_items', index=2,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=303,
  serialized_end=424,
)

_EVALUATORCONFIG = _descriptor.Descriptor(
  name='EvaluatorConfig',
  full_name='rv.protos.EvaluatorConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='evaluator_type', full_name='rv.protos.EvaluatorConfig.evaluator_type', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='classification_config', full_name='rv.protos.EvaluatorConfig.classification_config', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='custom_config', full_name='rv.protos.EvaluatorConfig.custom_config', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_EVALUATORCONFIG_CLASSIFICATIONEVALUATORCONFIG, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='evaluator_config', full_name='rv.protos.EvaluatorConfig.evaluator_config',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=119,
  serialized_end=444,
)

_EVALUATORCONFIG_CLASSIFICATIONEVALUATORCONFIG.fields_by_name['class_items'].message_type = rastervision_dot_protos_dot_class__item__pb2._CLASSITEM
_EVALUATORCONFIG_CLASSIFICATIONEVALUATORCONFIG.containing_type = _EVALUATORCONFIG
_EVALUATORCONFIG.fields_by_name['classification_config'].message_type = _EVALUATORCONFIG_CLASSIFICATIONEVALUATORCONFIG
_EVALUATORCONFIG.fields_by_name['custom_config'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_EVALUATORCONFIG.oneofs_by_name['evaluator_config'].fields.append(
  _EVALUATORCONFIG.fields_by_name['classification_config'])
_EVALUATORCONFIG.fields_by_name['classification_config'].containing_oneof = _EVALUATORCONFIG.oneofs_by_name['evaluator_config']
_EVALUATORCONFIG.oneofs_by_name['evaluator_config'].fields.append(
  _EVALUATORCONFIG.fields_by_name['custom_config'])
_EVALUATORCONFIG.fields_by_name['custom_config'].containing_oneof = _EVALUATORCONFIG.oneofs_by_name['evaluator_config']
DESCRIPTOR.message_types_by_name['EvaluatorConfig'] = _EVALUATORCONFIG

EvaluatorConfig = _reflection.GeneratedProtocolMessageType('EvaluatorConfig', (_message.Message,), dict(

  ClassificationEvaluatorConfig = _reflection.GeneratedProtocolMessageType('ClassificationEvaluatorConfig', (_message.Message,), dict(
    DESCRIPTOR = _EVALUATORCONFIG_CLASSIFICATIONEVALUATORCONFIG,
    __module__ = 'rastervision.protos.evaluator_pb2'
    # @@protoc_insertion_point(class_scope:rv.protos.EvaluatorConfig.ClassificationEvaluatorConfig)
    ))
  ,
  DESCRIPTOR = _EVALUATORCONFIG,
  __module__ = 'rastervision.protos.evaluator_pb2'
  # @@protoc_insertion_point(class_scope:rv.protos.EvaluatorConfig)
  ))
_sym_db.RegisterMessage(EvaluatorConfig)
_sym_db.RegisterMessage(EvaluatorConfig.ClassificationEvaluatorConfig)


# @@protoc_insertion_point(module_scope)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/augmentor.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/augmentor.proto',
  package='rv.protos',
  syntax='proto2',
  serialized_pb=_b('\n#rastervision/protos/augmentor.proto\x12\trv.protos\x1a\x1cgoogle/protobuf/struct.proto\"\x83\x01\n\x0f\x41ugmentorConfig\x12\x16\n\x0e\x61ugmentor_type\x18\x01 \x02(\t\x12\x12\n\x08\x61ug_prob\x18\x03 \x01(\x02H\x00\x12\x30\n\rcustom_config\x18\x05 \x01(\x0b\x32\x17.google.protobuf.StructH\x00\x42\x12\n\x10\x61ugmentor_config')
  ,
  dependencies=[google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_AUGMENTORCONFIG = _descriptor.Descriptor(
  name='AugmentorConfig',
  full_name='rv.protos.AugmentorConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='augmentor_type', full_name='rv.protos.AugmentorConfig.augmentor_type', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='aug_prob', full_name='rv.protos.AugmentorConfig.aug_prob', index=1,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='custom_config', full_name='rv.protos.AugmentorConfig.custom_config', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='augmentor_config', full_name='rv.protos.AugmentorConfig.augmentor_config',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=81,
  serialized_end=212,
)

_AUGMENTORCONFIG.fields_by_name['custom_config'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_AUGMENTORCONFIG.oneofs_by_name['augmentor_config'].fields.append(
  _AUGMENTORCONFIG.fields_by_name['aug_prob'])
_AUGMENTORCONFIG.fields_by_name['aug_prob'].containing_oneof = _AUGMENTORCONFIG.oneofs_by_name['augmentor_config']
_AUGMENTORCONFIG.oneofs_by_name['augmentor_config'].fields.append(
  _AUGMENTORCONFIG.fields_by_name['custom_config'])
_AUGMENTORCONFIG.fields_by_name['custom_config'].containing_oneof = _AUGMENTORCONFIG.oneofs_by_name['augmentor_config']
DESCRIPTOR.message_types_by_name['AugmentorConfig'] = _AUGMENTORCONFIG

AugmentorConfig = _reflection.GeneratedProtocolMessageType('AugmentorConfig', (_message.Message,), dict(
  DESCRIPTOR = _AUGMENTORCONFIG,
  __module__ = 'rastervision.protos.augmentor_pb2'
  # @@protoc_insertion_point(class_scope:rv.protos.AugmentorConfig)
  ))
_sym_db.RegisterMessage(AugmentorConfig)


# @@protoc_insertion_point(module_scope)



from copy import deepcopy

import rastervision as rv
from rastervision.command import (ChipCommand, CommandConfig,
                                  CommandConfigBuilder, NoOpCommand)
from rastervision.protos.command_pb2 \
    import CommandConfig as CommandConfigMsg
from rastervision.rv_config import RVConfig
from rastervision.data import SceneConfig
from rastervision.command.utils import (check_task_type, check_backend_type)
from rastervision.utils.misc import split_into_groups


class ChipCommandConfig(CommandConfig):
    def __init__(self, root_uri, split_id, task, backend, augmentors,
                 train_scenes, val_scenes):
        super().__init__(rv.CHIP, root_uri, split_id)
        self.task = task
        self.backend = backend
        self.augmentors = augmentors
        self.train_scenes = train_scenes
        self.val_scenes = val_scenes

    def create_command(self, tmp_dir=None):
        if len(self.train_scenes) == 0 and len(self.val_scenes) == 0:
            return NoOpCommand()

        if not tmp_dir:
            _tmp_dir = RVConfig.get_tmp_dir()
            tmp_dir = _tmp_dir.name
        else:
            _tmp_dir = tmp_dir

        retval = ChipCommand(self)
        retval.set_tmp_dir(_tmp_dir)
        return retval

    def to_proto(self):
        msg = super().to_proto()

        task = self.task.to_proto()
        backend = self.backend.to_proto()
        train_scenes = list(map(lambda s: s.to_proto(), self.train_scenes))
        val_scenes = list(map(lambda s: s.to_proto(), self.val_scenes))
        msg.MergeFrom(
            CommandConfigMsg(
                chip_config=CommandConfigMsg.ChipConfig(
                    task=task,
                    backend=backend,
                    train_scenes=train_scenes,
                    val_scenes=val_scenes)))

        return msg

    def report_io(self):
        io_def = rv.core.CommandIODefinition()
        self.task.report_io(self.command_type, io_def)
        self.backend.report_io(self.command_type, io_def)
        for scene in self.train_scenes:
            scene.report_io(self.command_type, io_def)
        for scene in self.val_scenes:
            scene.report_io(self.command_type, io_def)
        for augmentor in self.augmentors:
            augmentor.report_io(self.command_type, io_def)
        return io_def

    def split(self, num_parts):
        commands = []
        t_scenes = list(map(lambda x: (0, x), self.train_scenes))
        v_scenes = list(map(lambda x: (1, x), self.val_scenes))

        for i, l in enumerate(
                split_into_groups(t_scenes + v_scenes, num_parts)):
            split_t_scenes = list(
                map(lambda x: x[1], filter(lambda x: x[0] == 0, l)))
            split_v_scenes = list(
                map(lambda x: x[1], filter(lambda x: x[0] == 1, l)))
            c = self.to_builder() \
                .with_train_scenes(split_t_scenes) \
                .with_val_scenes(split_v_scenes) \
                .with_split_id(i) \
                .build()
            commands.append(c)
        return commands


class ChipCommandConfigBuilder(CommandConfigBuilder):
    def __init__(self, command_type, prev=None):
        super().__init__(command_type, prev)
        if prev is None:
            self.task = None
            self.backend = None
            self.augmentors = []
            self.train_scenes = []
            self.val_scenes = []
        else:
            self.task = prev.task
            self.backend = prev.backend
            self.augmentors = prev.augmentors
            self.train_scenes = prev.train_scenes
            self.val_scenes = prev.val_scenes

    def validate(self):
        super().validate()
        if self.task is None:
            raise rv.ConfigError('Task not set for ChipCommandConfig. Use '
                                 'with_task or with_experiment')
        check_task_type(self.task)
        if self.backend is None:
            raise rv.ConfigError('Backend not set for ChipCommandConfig. Use '
                                 'with_backend or with_experiment')
        check_backend_type(self.backend)
        if len(self.train_scenes) > 0:
            for s in self.train_scenes:
                if not isinstance(s, SceneConfig):
                    raise rv.ConfigError(
                        'train_scenes must be a list of class SceneConfig, '
                        'got a list of {}'.format(type(s)))
        if len(self.val_scenes) > 0:
            for s in self.val_scenes:
                if not isinstance(s, SceneConfig):
                    raise rv.ConfigError(
                        'val_scenes must be a list of class SceneConfig, '
                        'got a list of {}'.format(type(s)))

    def build(self):
        self.validate()
        return ChipCommandConfig(self.root_uri, self.split_id, self.task,
                                 self.backend, self.augmentors,
                                 self.train_scenes, self.val_scenes)

    def from_proto(self, msg):
        b = super().from_proto(msg)

        conf = msg.chip_config

        task = rv.TaskConfig.from_proto(conf.task)
        backend = rv.BackendConfig.from_proto(conf.backend)
        augmentors = list(map(rv.AugmentorConfig.from_proto, conf.augmentors))
        train_scenes = list(map(rv.SceneConfig.from_proto, conf.train_scenes))
        val_scenes = list(map(rv.SceneConfig.from_proto, conf.val_scenes))

        b = b.with_task(task)
        b = b.with_backend(backend)
        b = b.with_augmentors(augmentors)
        b = b.with_train_scenes(train_scenes)
        b = b.with_val_scenes(val_scenes)

        return b

    def get_root_uri(self, experiment_config):
        return experiment_config.chip_uri

    def with_experiment(self, experiment_config):
        b = super().with_experiment(experiment_config)
        b = b.with_task(experiment_config.task)
        b = b.with_backend(experiment_config.backend)
        b = b.with_augmentors(experiment_config.dataset.augmentors)
        b = b.with_train_scenes(experiment_config.dataset.train_scenes)
        b = b.with_val_scenes(experiment_config.dataset.validation_scenes)
        return b

    def with_task(self, task):
        """Sets a specific task type.

        Args:
            task:  A TaskConfig object.

        """
        b = deepcopy(self)
        b.task = task
        return b

    def with_backend(self, backend):
        b = deepcopy(self)
        b.backend = backend
        return b

    def with_augmentors(self, augmentors):
        b = deepcopy(self)
        b.augmentors = augmentors
        return b

    def with_train_scenes(self, scenes):
        b = deepcopy(self)
        b.train_scenes = scenes
        return b

    def with_val_scenes(self, scenes):
        b = deepcopy(self)
        b.val_scenes = scenes
        return b



# flake8: noqa

# Registry Keys

RASTER_TRANSFORMER = 'RASTER_TRANSFORMER'

CHANNEL_TRANSFORMER = 'CHANNEL_TRANSFORMER'
STATS_TRANSFORMER = 'STATS_TRANSFORMER'

from rastervision.data.raster_transformer.raster_transformer_config \
    import RasterTransformerConfig



from copy import deepcopy

import rastervision as rv
from rastervision.command import (AnalyzeCommand, CommandConfig,
                                  CommandConfigBuilder, NoOpCommand)
from rastervision.protos.command_pb2 \
    import CommandConfig as CommandConfigMsg
from rastervision.command.utils import (check_scenes_type,
                                        check_analyzers_type)


class AnalyzeCommandConfig(CommandConfig):
    def __init__(self, root_uri, task, scenes, analyzers):
        super().__init__(rv.ANALYZE, root_uri)
        self.task = task
        self.scenes = scenes
        self.analyzers = analyzers

    def create_command(self, tmp_dir=None):
        if len(self.scenes) == 0 or len(self.analyzers) == 0:
            return NoOpCommand()

        retval = AnalyzeCommand(self)
        retval.set_tmp_dir(tmp_dir)
        return retval

    def to_proto(self):
        msg = super().to_proto()
        task = self.task.to_proto()
        scenes = list(map(lambda s: s.to_proto(), self.scenes))
        analyzers = list(map(lambda a: a.to_proto(), self.analyzers))

        msg.MergeFrom(
            CommandConfigMsg(
                analyze_config=CommandConfigMsg.AnalyzeConfig(
                    task=task, scenes=scenes, analyzers=analyzers)))

        return msg

    def report_io(self):
        io_def = rv.core.CommandIODefinition()
        self.task.report_io(self.command_type, io_def)
        for scene in self.scenes:
            scene.report_io(self.command_type, io_def)
        for analyzer in self.analyzers:
            analyzer.report_io(self.command_type, io_def)
        return io_def


class AnalyzeCommandConfigBuilder(CommandConfigBuilder):
    def __init__(self, command_type, prev=None):
        super().__init__(command_type, prev)
        if prev is None:
            self.task = None
            self.scenes = None
            self.analyzers = None
        else:
            self.task = prev.task
            self.scenes = prev.scenes
            self.analyzers = prev.analyzers

    def validate(self):
        super().validate()
        if self.scenes is None:
            raise rv.ConfigError('scenes not set for AnalyzeCommandConfig. Use'
                                 ' with_scenes or with_experiment')
        check_scenes_type(self.scenes)
        if self.analyzers is None:
            raise rv.ConfigError(
                'analyzers not set. Use with_analyzers or with_experiment')
        check_analyzers_type(self.analyzers)

    def build(self):
        self.validate()
        return AnalyzeCommandConfig(self.root_uri, self.task, self.scenes,
                                    self.analyzers)

    def from_proto(self, msg):
        b = super().from_proto(msg)

        conf = msg.analyze_config

        task = rv.TaskConfig.from_proto(conf.task)
        scenes = list(map(rv.SceneConfig.from_proto, conf.scenes))
        analyzers = list(map(rv.AnalyzerConfig.from_proto, conf.analyzers))

        b = b.with_task(task)
        b = b.with_scenes(scenes)
        b = b.with_analyzers(analyzers)

        return b

    def get_root_uri(self, experiment_config):
        return experiment_config.analyze_uri

    def with_experiment(self, experiment_config):
        b = super().with_experiment(experiment_config)
        b = b.with_task(experiment_config.task)
        b = b.with_scenes(experiment_config.dataset.all_scenes())
        b = b.with_analyzers(experiment_config.analyzers)
        return b

    def with_task(self, task):
        """Sets a specific task type.

        Args:
            task:  A TaskConfig object.

        """
        b = deepcopy(self)
        b.task = task
        return b

    def with_scenes(self, scenes):
        b = deepcopy(self)
        b.scenes = scenes
        return b

    def with_analyzers(self, analyzers):
        b = deepcopy(self)
        b.analyzers = analyzers
        return b



from os.path import join

import numpy as np
from PIL import Image, ImageDraw

from rastervision.rv_config import RVConfig
from rastervision.core import Box
from rastervision.task import Task
from rastervision.utils.files import (get_local_path, upload_or_copy, make_dir)


def draw_debug_predict_image(scene, class_map):
    img = scene.raster_source.get_image_array()
    img = Image.fromarray(img)
    draw = ImageDraw.Draw(img, 'RGB')
    labels = scene.prediction_label_store.get_labels()
    line_width = 4
    default_colors = [
        'red', 'orange', 'yellow', 'green', 'brown', 'pink', 'purple'
    ]
    for cell, class_id in zip(labels.get_cells(), labels.get_class_ids()):
        cell = cell.make_eroded(line_width // 2)
        coords = cell.geojson_coordinates()
        color = class_map.get_by_id(class_id).color
        if color is None:
            color = default_colors[(class_id - 1) % len(default_colors)]
        draw.line(coords, fill=color, width=line_width)
    return img


class ChipClassification(Task):
    def get_train_windows(self, scene):
        result = []
        extent = scene.raster_source.get_extent()
        chip_size = self.config.chip_size
        stride = chip_size
        windows = extent.get_windows(chip_size, stride)
        if scene.aoi_polygons:
            windows = Box.filter_by_aoi(windows, scene.aoi_polygons)
        for window in windows:
            chip = scene.raster_source.get_chip(window)
            if np.sum(chip.ravel()) > 0:
                result.append(window)
        return result

    def get_train_labels(self, window, scene):
        return scene.ground_truth_label_source.get_labels(window=window)

    def post_process_predictions(self, labels, scene):
        return labels

    def get_predict_windows(self, extent):
        chip_size = self.config.chip_size
        stride = chip_size
        return extent.get_windows(chip_size, stride)

    def save_debug_predict_image(self, scene, debug_dir_uri):
        img = draw_debug_predict_image(scene, self.config.class_map)
        # Saving to a jpg leads to segfault for unknown reasons.
        debug_image_uri = join(debug_dir_uri, scene.id + '.png')
        with RVConfig.get_tmp_dir() as temp_dir:
            debug_image_path = get_local_path(debug_image_uri, temp_dir)
            make_dir(debug_image_path, use_dirname=True)
            img.save(debug_image_path)
            upload_or_copy(debug_image_path, debug_image_uri)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/label_store.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/label_store.proto',
  package='rv.protos',
  syntax='proto2',
  serialized_pb=_b('\n%rastervision/protos/label_store.proto\x12\trv.protos\x1a\x1cgoogle/protobuf/struct.proto\"\xd5\x05\n\x10LabelStoreConfig\x12\x12\n\nstore_type\x18\x01 \x02(\t\x12\r\n\x03uri\x18\x02 \x01(\tH\x00\x12i\n\"semantic_segmentation_raster_store\x18\x03 \x01(\x0b\x32;.rv.protos.LabelStoreConfig.SemanticSegmentationRasterStoreH\x00\x12\x30\n\rcustom_config\x18\x04 \x01(\x0b\x32\x17.google.protobuf.StructH\x00\x1a\xea\x03\n\x1fSemanticSegmentationRasterStore\x12\x0b\n\x03uri\x18\x01 \x01(\t\x12\x12\n\x03rgb\x18\x02 \x01(\x08:\x05\x66\x61lse\x12_\n\rvector_output\x18\x03 \x03(\x0b\x32H.rv.protos.LabelStoreConfig.SemanticSegmentationRasterStore.VectorOutput\x1a\x89\x01\n\x0f\x42uildingOptions\x12\x1f\n\x10min_aspect_ratio\x18\x01 \x01(\x02:\x05\x31.618\x12\x10\n\x08min_area\x18\x02 \x01(\x02\x12!\n\x14\x65lement_width_factor\x18\x03 \x01(\x02:\x03\x30.5\x12 \n\x11\x65lement_thickness\x18\x04 \x01(\x02:\x05\x30.001\x1a\xb8\x01\n\x0cVectorOutput\x12\x12\n\x07\x64\x65noise\x18\x01 \x01(\x05:\x01\x30\x12\r\n\x03uri\x18\x02 \x01(\t:\x00\x12\x0c\n\x04mode\x18\x03 \x02(\t\x12\x10\n\x08\x63lass_id\x18\x04 \x02(\x05\x12\x65\n\x10\x62uilding_options\x18\x05 \x01(\x0b\x32K.rv.protos.LabelStoreConfig.SemanticSegmentationRasterStore.BuildingOptionsB\x14\n\x12label_store_config')
  ,
  dependencies=[google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_LABELSTORECONFIG_SEMANTICSEGMENTATIONRASTERSTORE_BUILDINGOPTIONS = _descriptor.Descriptor(
  name='BuildingOptions',
  full_name='rv.protos.LabelStoreConfig.SemanticSegmentationRasterStore.BuildingOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='min_aspect_ratio', full_name='rv.protos.LabelStoreConfig.SemanticSegmentationRasterStore.BuildingOptions.min_aspect_ratio', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1.618),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='min_area', full_name='rv.protos.LabelStoreConfig.SemanticSegmentationRasterStore.BuildingOptions.min_area', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='element_width_factor', full_name='rv.protos.LabelStoreConfig.SemanticSegmentationRasterStore.BuildingOptions.element_width_factor', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.5),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='element_thickness', full_name='rv.protos.LabelStoreConfig.SemanticSegmentationRasterStore.BuildingOptions.element_thickness', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.001),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=462,
  serialized_end=599,
)

_LABELSTORECONFIG_SEMANTICSEGMENTATIONRASTERSTORE_VECTOROUTPUT = _descriptor.Descriptor(
  name='VectorOutput',
  full_name='rv.protos.LabelStoreConfig.SemanticSegmentationRasterStore.VectorOutput',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='denoise', full_name='rv.protos.LabelStoreConfig.SemanticSegmentationRasterStore.VectorOutput.denoise', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='uri', full_name='rv.protos.LabelStoreConfig.SemanticSegmentationRasterStore.VectorOutput.uri', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='mode', full_name='rv.protos.LabelStoreConfig.SemanticSegmentationRasterStore.VectorOutput.mode', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='class_id', full_name='rv.protos.LabelStoreConfig.SemanticSegmentationRasterStore.VectorOutput.class_id', index=3,
      number=4, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='building_options', full_name='rv.protos.LabelStoreConfig.SemanticSegmentationRasterStore.VectorOutput.building_options', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=602,
  serialized_end=786,
)

_LABELSTORECONFIG_SEMANTICSEGMENTATIONRASTERSTORE = _descriptor.Descriptor(
  name='SemanticSegmentationRasterStore',
  full_name='rv.protos.LabelStoreConfig.SemanticSegmentationRasterStore',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='uri', full_name='rv.protos.LabelStoreConfig.SemanticSegmentationRasterStore.uri', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rgb', full_name='rv.protos.LabelStoreConfig.SemanticSegmentationRasterStore.rgb', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='vector_output', full_name='rv.protos.LabelStoreConfig.SemanticSegmentationRasterStore.vector_output', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_LABELSTORECONFIG_SEMANTICSEGMENTATIONRASTERSTORE_BUILDINGOPTIONS, _LABELSTORECONFIG_SEMANTICSEGMENTATIONRASTERSTORE_VECTOROUTPUT, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=296,
  serialized_end=786,
)

_LABELSTORECONFIG = _descriptor.Descriptor(
  name='LabelStoreConfig',
  full_name='rv.protos.LabelStoreConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='store_type', full_name='rv.protos.LabelStoreConfig.store_type', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='uri', full_name='rv.protos.LabelStoreConfig.uri', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='semantic_segmentation_raster_store', full_name='rv.protos.LabelStoreConfig.semantic_segmentation_raster_store', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='custom_config', full_name='rv.protos.LabelStoreConfig.custom_config', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_LABELSTORECONFIG_SEMANTICSEGMENTATIONRASTERSTORE, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='label_store_config', full_name='rv.protos.LabelStoreConfig.label_store_config',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=83,
  serialized_end=808,
)

_LABELSTORECONFIG_SEMANTICSEGMENTATIONRASTERSTORE_BUILDINGOPTIONS.containing_type = _LABELSTORECONFIG_SEMANTICSEGMENTATIONRASTERSTORE
_LABELSTORECONFIG_SEMANTICSEGMENTATIONRASTERSTORE_VECTOROUTPUT.fields_by_name['building_options'].message_type = _LABELSTORECONFIG_SEMANTICSEGMENTATIONRASTERSTORE_BUILDINGOPTIONS
_LABELSTORECONFIG_SEMANTICSEGMENTATIONRASTERSTORE_VECTOROUTPUT.containing_type = _LABELSTORECONFIG_SEMANTICSEGMENTATIONRASTERSTORE
_LABELSTORECONFIG_SEMANTICSEGMENTATIONRASTERSTORE.fields_by_name['vector_output'].message_type = _LABELSTORECONFIG_SEMANTICSEGMENTATIONRASTERSTORE_VECTOROUTPUT
_LABELSTORECONFIG_SEMANTICSEGMENTATIONRASTERSTORE.containing_type = _LABELSTORECONFIG
_LABELSTORECONFIG.fields_by_name['semantic_segmentation_raster_store'].message_type = _LABELSTORECONFIG_SEMANTICSEGMENTATIONRASTERSTORE
_LABELSTORECONFIG.fields_by_name['custom_config'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_LABELSTORECONFIG.oneofs_by_name['label_store_config'].fields.append(
  _LABELSTORECONFIG.fields_by_name['uri'])
_LABELSTORECONFIG.fields_by_name['uri'].containing_oneof = _LABELSTORECONFIG.oneofs_by_name['label_store_config']
_LABELSTORECONFIG.oneofs_by_name['label_store_config'].fields.append(
  _LABELSTORECONFIG.fields_by_name['semantic_segmentation_raster_store'])
_LABELSTORECONFIG.fields_by_name['semantic_segmentation_raster_store'].containing_oneof = _LABELSTORECONFIG.oneofs_by_name['label_store_config']
_LABELSTORECONFIG.oneofs_by_name['label_store_config'].fields.append(
  _LABELSTORECONFIG.fields_by_name['custom_config'])
_LABELSTORECONFIG.fields_by_name['custom_config'].containing_oneof = _LABELSTORECONFIG.oneofs_by_name['label_store_config']
DESCRIPTOR.message_types_by_name['LabelStoreConfig'] = _LABELSTORECONFIG

LabelStoreConfig = _reflection.GeneratedProtocolMessageType('LabelStoreConfig', (_message.Message,), dict(

  SemanticSegmentationRasterStore = _reflection.GeneratedProtocolMessageType('SemanticSegmentationRasterStore', (_message.Message,), dict(

    BuildingOptions = _reflection.GeneratedProtocolMessageType('BuildingOptions', (_message.Message,), dict(
      DESCRIPTOR = _LABELSTORECONFIG_SEMANTICSEGMENTATIONRASTERSTORE_BUILDINGOPTIONS,
      __module__ = 'rastervision.protos.label_store_pb2'
      # @@protoc_insertion_point(class_scope:rv.protos.LabelStoreConfig.SemanticSegmentationRasterStore.BuildingOptions)
      ))
    ,

    VectorOutput = _reflection.GeneratedProtocolMessageType('VectorOutput', (_message.Message,), dict(
      DESCRIPTOR = _LABELSTORECONFIG_SEMANTICSEGMENTATIONRASTERSTORE_VECTOROUTPUT,
      __module__ = 'rastervision.protos.label_store_pb2'
      # @@protoc_insertion_point(class_scope:rv.protos.LabelStoreConfig.SemanticSegmentationRasterStore.VectorOutput)
      ))
    ,
    DESCRIPTOR = _LABELSTORECONFIG_SEMANTICSEGMENTATIONRASTERSTORE,
    __module__ = 'rastervision.protos.label_store_pb2'
    # @@protoc_insertion_point(class_scope:rv.protos.LabelStoreConfig.SemanticSegmentationRasterStore)
    ))
  ,
  DESCRIPTOR = _LABELSTORECONFIG,
  __module__ = 'rastervision.protos.label_store_pb2'
  # @@protoc_insertion_point(class_scope:rv.protos.LabelStoreConfig)
  ))
_sym_db.RegisterMessage(LabelStoreConfig)
_sym_db.RegisterMessage(LabelStoreConfig.SemanticSegmentationRasterStore)
_sym_db.RegisterMessage(LabelStoreConfig.SemanticSegmentationRasterStore.BuildingOptions)
_sym_db.RegisterMessage(LabelStoreConfig.SemanticSegmentationRasterStore.VectorOutput)


# @@protoc_insertion_point(module_scope)



import tempfile

from PIL import Image
import numpy as np
import click
import mercantile
import rasterio
from rasterio.windows import Window
import pyproj

from rastervision.utils.files import (download_if_needed, get_local_path,
                                      upload_or_copy)
from rastervision.command.aux.cogify_command import create_cog


def lnglat2merc(lng, lat):
    """Convert lng, lat point to x/y Web Mercator tuple."""
    return pyproj.transform(
        pyproj.Proj(init='epsg:4326'), pyproj.Proj(init='epsg:3857'), lng, lat)


def merc2lnglat(x, y):
    """Convert x, y Web Mercator point to lng/lat tuple."""
    return pyproj.transform(
        pyproj.Proj(init='epsg:3857'), pyproj.Proj(init='epsg:4326'), x, y)


def merc2pixel(tile_x, tile_y, zoom, merc_x, merc_y, tile_sz=256):
    """Convert Web Mercator point to pixel coordinates.

    This is within the coordinate frame of a single ZXY tile.

    Args:
        tile_x: (int) x coordinate of ZXY tile
        tile_y: (int) y coordinate of ZXY tile
        zoom: (int) zoom level of ZXY tile
        merc_x: (float) Web Mercator x axis of point
        merc_y: (float) Web Mercator y axis of point
        tile_sz: (int) size of ZXY tile
    """
    tile_merc_bounds = mercantile.xy_bounds(tile_x, tile_y, zoom)
    pix_y = int(
        round(tile_sz * ((tile_merc_bounds.top - merc_y) /
                         (tile_merc_bounds.top - tile_merc_bounds.bottom))))
    pix_x = int(
        round(tile_sz * ((merc_x - tile_merc_bounds.left) /
                         (tile_merc_bounds.right - tile_merc_bounds.left))))
    return (pix_x, pix_y)


def _zxy2geotiff(tile_schema, zoom, bounds, output_uri, make_cog=False):
    """Generates a GeoTIFF of a bounded region from a ZXY tile server.

    Args:
        tile_schema: (str) the URI schema for zxy tiles (ie. a slippy map tile server)
            of the form /tileserver-uri/{z}/{x}/{y}.png. If {-y} is used, the tiles
            are assumed to be indexed using TMS coordinates, where the y axis starts
            at the southernmost point. The URI can be for http, S3, or the local
            file system.
        zoom: (int) the zoom level to use when retrieving tiles
        bounds: (list) a list of length 4 containing min_lat, min_lng,
            max_lat, max_lng
        output_uri: (str) where to save the GeoTIFF. The URI can be for http, S3, or the
            local file system
    """
    min_lat, min_lng, max_lat, max_lng = bounds
    if min_lat >= max_lat:
        raise ValueError('min_lat must be < max_lat')
    if min_lng >= max_lng:
        raise ValueError('min_lng must be < max_lng')

    is_tms = False
    if '{-y}' in tile_schema:
        tile_schema = tile_schema.replace('{-y}', '{y}')
        is_tms = True

    tmp_dir_obj = tempfile.TemporaryDirectory()
    tmp_dir = tmp_dir_obj.name

    # Get range of tiles that cover bounds.
    output_path = get_local_path(output_uri, tmp_dir)
    tile_sz = 256
    t = mercantile.tile(min_lng, max_lat, zoom)
    xmin, ymin = t.x, t.y
    t = mercantile.tile(max_lng, min_lat, zoom)
    xmax, ymax = t.x, t.y

    # The supplied bounds are contained within the "tile bounds" -- ie. the
    # bounds of the set of tiles that covers the supplied bounds. Therefore,
    # we need to crop out the imagery that lies within the supplied bounds.
    # We do this by computing a top, bottom, left, and right offset in pixel
    # units of the supplied bounds against the tile bounds. Getting the offsets
    # in pixel units involves converting lng/lat to web mercator units since we
    # assume that is the CRS of the tiles. These offsets are then used to crop
    # individual tiles and place them correctly into the output raster.
    nw_merc_x, nw_merc_y = lnglat2merc(min_lng, max_lat)
    left_pix_offset, top_pix_offset = merc2pixel(xmin, ymin, zoom, nw_merc_x,
                                                 nw_merc_y)

    se_merc_x, se_merc_y = lnglat2merc(max_lng, min_lat)
    se_left_pix_offset, se_top_pix_offset = merc2pixel(xmax, ymax, zoom,
                                                       se_merc_x, se_merc_y)
    right_pix_offset = tile_sz - se_left_pix_offset
    bottom_pix_offset = tile_sz - se_top_pix_offset

    uncropped_height = tile_sz * (ymax - ymin + 1)
    uncropped_width = tile_sz * (xmax - xmin + 1)
    height = uncropped_height - top_pix_offset - bottom_pix_offset
    width = uncropped_width - left_pix_offset - right_pix_offset

    transform = rasterio.transform.from_bounds(nw_merc_x, se_merc_y, se_merc_x,
                                               nw_merc_y, width, height)
    with rasterio.open(
            output_path,
            'w',
            driver='GTiff',
            height=height,
            width=width,
            count=3,
            crs='epsg:3857',
            transform=transform,
            dtype=rasterio.uint8) as dataset:
        out_x = 0
        for xi, x in enumerate(range(xmin, xmax + 1)):
            tile_xmin, tile_xmax = 0, tile_sz - 1
            if x == xmin:
                tile_xmin += left_pix_offset
            if x == xmax:
                tile_xmax -= right_pix_offset
            window_width = tile_xmax - tile_xmin + 1

            out_y = 0
            for yi, y in enumerate(range(ymin, ymax + 1)):
                tile_ymin, tile_ymax = 0, tile_sz - 1
                if y == ymin:
                    tile_ymin += top_pix_offset
                if y == ymax:
                    tile_ymax -= bottom_pix_offset
                window_height = tile_ymax - tile_ymin + 1

                # Convert from xyz to tms if needed.
                # https://gist.github.com/tmcw/4954720
                if is_tms:
                    y = (2**zoom) - y - 1
                tile_uri = tile_schema.format(x=x, y=y, z=zoom)
                tile_path = download_if_needed(tile_uri, tmp_dir)
                img = np.array(Image.open(tile_path))
                img = img[tile_ymin:tile_ymax + 1, tile_xmin:tile_xmax + 1, :]

                window = Window(out_x, out_y, window_width, window_height)
                dataset.write(
                    np.transpose(img[:, :, 0:3], (2, 0, 1)), window=window)
                out_y += window_height
            out_x += window_width

    if make_cog:
        create_cog(output_path, output_uri, tmp_dir)
    else:
        upload_or_copy(output_path, output_uri)


@click.command()
@click.argument('tile_schema')
@click.argument('zoom')
@click.argument('bounds')
@click.argument('output_uri')
@click.option('--make-cog', is_flag=True, default=False)
def zxy2geotiff(tile_schema, zoom, bounds, output_uri, make_cog):
    """Generates a GeoTIFF of a bounded region from a ZXY tile server.

    TILE_SCHEMA: the URI schema for zxy tiles (ie. a slippy map tile server) of
    the form /tileserver-uri/{z}/{x}/{y}.png. If {-y} is used, the tiles are
    assumed to be indexed using TMS coordinates, where the y axis starts at
    the southernmost point. The URI can be for http, S3, or the local file
    system.

    ZOOM: the zoom level to use when retrieving tiles

    BOUNDS: a space-separated string containing min_lat, min_lng, max_lat,
    max_lng

    OUTPUT_URI: where to save the GeoTIFF. The URI can be for http, S3, or the
    local file system.
    """
    bounds = [float(x) for x in bounds.split(' ')]
    _zxy2geotiff(tile_schema, int(zoom), bounds, output_uri, make_cog=make_cog)


if __name__ == '__main__':
    zxy2geotiff()



import os
import shutil
from datetime import datetime, timezone
import glob

from rastervision.filesystem import (FileSystem, NotReadableError)


def make_dir(path, check_empty=False, force_empty=False, use_dirname=False):
    """Make a local directory.

    Args:
        path: path to directory
        check_empty: if True, check that directory is empty
        force_empty: if True, delete files if necessary to make directory
            empty
        use_dirname: if path is a file, use the the parent directory as path

    Raises:
        ValueError if check_empty is True and directory is not empty
    """
    directory = path
    if use_dirname:
        directory = os.path.abspath(os.path.dirname(path))

    if force_empty and os.path.isdir(directory):
        shutil.rmtree(directory)

    os.makedirs(directory, exist_ok=True)

    if check_empty and any(os.scandir(directory)):
        raise ValueError(
            '{} needs to be an empty directory!'.format(directory))


class LocalFileSystem(FileSystem):
    @staticmethod
    def matches_uri(uri: str, mode: str) -> bool:
        return True

    @staticmethod
    def file_exists(uri: str, include_dir: bool = True) -> bool:
        return (os.path.isfile(uri) or (include_dir and os.path.isdir(uri)))

    @staticmethod
    def read_str(file_uri: str) -> str:
        if not os.path.isfile(file_uri):
            raise NotReadableError('Could not read {}'.format(file_uri))
        with open(file_uri, 'r') as file_buffer:
            return file_buffer.read()

    @staticmethod
    def read_bytes(file_uri: str) -> bytes:
        if not os.path.isfile(file_uri):
            raise NotReadableError('Could not read {}'.format(file_uri))
        with open(file_uri, 'rb') as file_buffer:
            return file_buffer.read()

    @staticmethod
    def write_str(file_uri: str, data: str) -> None:
        make_dir(file_uri, use_dirname=True)
        with open(file_uri, 'w') as content_file:
            content_file.write(data)

    @staticmethod
    def write_bytes(file_uri: str, data: bytes) -> None:
        make_dir(file_uri, use_dirname=True)
        with open(file_uri, 'wb') as content_file:
            content_file.write(data)

    @staticmethod
    def sync_from_dir(src_dir_uri: str,
                      dest_dir_uri: str,
                      delete: bool = False) -> None:
        if src_dir_uri == dest_dir_uri:
            return

        if delete:
            shutil.rmtree(dest_dir_uri)

        # https://stackoverflow.com/a/15824216/841563
        def recursive_overwrite(src, dest):
            if os.path.isdir(src):
                if not os.path.isdir(dest):
                    os.makedirs(dest)
                    for entry in os.scandir(src):
                        recursive_overwrite(entry.path,
                                            os.path.join(dest, entry.name))
            else:
                shutil.copyfile(src, dest)

        recursive_overwrite(src_dir_uri, dest_dir_uri)

    @staticmethod
    def sync_to_dir(src_dir_uri: str, dest_dir_uri: str,
                    delete: bool = False) -> None:
        LocalFileSystem.sync_from_dir(src_dir_uri, dest_dir_uri, delete)

    @staticmethod
    def copy_to(src_path: str, dst_uri: str) -> None:
        if src_path != dst_uri:
            make_dir(dst_uri, use_dirname=True)
            shutil.copyfile(src_path, dst_uri)

    @staticmethod
    def copy_from(uri: str, path: str) -> None:
        not_found = not os.path.isfile(path)
        if not_found:
            raise NotReadableError('Could not read {}'.format(uri))

    @staticmethod
    def local_path(uri: str, download_dir: str) -> None:
        path = uri
        return path

    @staticmethod
    def last_modified(uri: str) -> datetime:
        local_last_modified = datetime.utcfromtimestamp(os.path.getmtime(uri))
        return local_last_modified.replace(tzinfo=timezone.utc)

    @staticmethod
    def list_paths(uri, ext=None):
        if ext is None:
            ext = ''
        return glob.glob(os.path.join(uri, '*' + ext))



import logging

from rasterio.features import rasterize
import numpy as np
from shapely.geometry import shape
from shapely.strtree import STRtree
import shapely

from rastervision.core import Box
from rastervision.data import (ActivateMixin, ActivationError)
from rastervision.data.raster_source import RasterSource

log = logging.getLogger(__name__)


def geoms_to_raster(str_tree, rasterizer_options, window, extent):
    background_class_id = rasterizer_options.background_class_id
    all_touched = rasterizer_options.all_touched

    log.debug('Cropping shapes to window...')
    # Crop shapes against window, remove empty shapes, and put in window frame of
    # reference.
    window_geom = window.to_shapely()
    shapes = str_tree.query(window_geom)
    shapes = [(s, s.class_id) for s in shapes]
    shapes = [(s.intersection(window_geom), c) for s, c in shapes]
    shapes = [(s, c) for s, c in shapes if not s.is_empty]

    def to_window_frame(x, y, z=None):
        return (x - window.xmin, y - window.ymin)

    shapes = [(shapely.ops.transform(to_window_frame, s), c)
              for s, c in shapes]
    log.debug('# of shapes in window: {}'.format(len(shapes)))

    out_shape = (window.get_height(), window.get_width())

    # rasterize needs to be passed >= 1 shapes.
    if shapes:
        log.debug('rasterio.rasterize()...')
        raster = rasterize(
            shapes,
            out_shape=out_shape,
            fill=background_class_id,
            dtype=np.uint8,
            all_touched=all_touched)
    else:
        raster = np.full(out_shape, background_class_id, dtype=np.uint8)

    # Ensure that parts of window outside of extent have zero values which are counted as
    # the don't-care class for segmentation.
    valid_window = window_geom.intersection(extent.to_shapely())
    if valid_window.is_empty:
        raster[:, :] = 0
    else:
        vw = shapely.ops.transform(to_window_frame, valid_window)
        vw = Box.from_shapely(vw).to_int()
        new_raster = np.zeros(out_shape)
        new_raster[vw.ymin:vw.ymax, vw.xmin:vw.xmax] = \
            raster[vw.ymin:vw.ymax, vw.xmin:vw.xmax]
        raster = new_raster

    return raster


class RasterizedSource(ActivateMixin, RasterSource):
    """A RasterSource based on the rasterization of a VectorSource."""

    def __init__(self, vector_source, rasterizer_options, extent,
                 crs_transformer):
        """Constructor.

        Args:
            vector_source: (VectorSource)
            rasterizer_options:
                rastervision.data.raster_source.GeoJSONSourceConfig.RasterizerOptions
            extent: (Box) extent of corresponding imagery RasterSource
            crs_transformer: (CRSTransformer)
        """
        self.vector_source = vector_source
        self.rasterizer_options = rasterizer_options
        self.extent = extent
        self.crs_transformer = crs_transformer
        self.activated = False

        super().__init__(channel_order=[0], num_channels=1)

    def get_extent(self):
        """Return the extent of the RasterSource.

        Returns:
            Box in pixel coordinates with extent
        """
        return self.extent

    def get_dtype(self):
        """Return the numpy.dtype of this scene"""
        return np.uint8

    def get_crs_transformer(self):
        """Return the associated CRSTransformer."""
        return self.crs_transformer

    def _get_chip(self, window):
        """Return the chip located in the window.

        Polygons falling within the window are rasterized using the class_id, and
        the background is filled with background_class_id. Also, any pixels in the
        window outside the extent are zero, which is the don't-care class for
        segmentation.

        Args:
            window: Box

        Returns:
            [height, width, channels] numpy array
        """
        if not self.activated:
            raise ActivationError('GeoJSONSource must be activated before use')

        log.debug('Rasterizing window: {}'.format(window))
        chip = geoms_to_raster(self.str_tree, self.rasterizer_options, window,
                               self.get_extent())
        # Add third singleton dim since rasters must have >=1 channel.
        return np.expand_dims(chip, 2)

    def _activate(self):
        geojson = self.vector_source.get_geojson()
        geoms = []
        for f in geojson['features']:
            geom = shape(f['geometry'])
            geom.class_id = f['properties']['class_id']
            geoms.append(geom)
        self.str_tree = STRtree(geoms)
        self.activated = True

    def _deactivate(self):
        self.str_tree = None
        self.activated = False



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/deeplab/train.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/deeplab/train.proto',
  package='deeplab.protos',
  syntax='proto2',
  serialized_pb=_b('\n\'rastervision/protos/deeplab/train.proto\x12\x0e\x64\x65\x65plab.protos\"\xeb\x07\n\x12TrainingParameters\x12\x1d\n\x15initialize_last_layer\x18\x01 \x01(\t\x12\'\n\x1flast_layers_contain_logits_only\x18\x02 \x01(\t\x12\x17\n\x0fupsample_logits\x18\x03 \x01(\t\x12\x1c\n\x14\x66ine_tune_batch_norm\x18\x04 \x01(\t\x12\x1a\n\x12\x62\x61se_learning_rate\x18\x05 \x01(\t\x12&\n\x1elast_layer_gradient_multiplier\x18\x06 \x01(\t\x12\x16\n\x0elearning_power\x18\x07 \x01(\t\x12\"\n\x1alearning_rate_decay_factor\x18\x08 \x01(\t\x12\x10\n\x08momentum\x18\t \x01(\t\x12 \n\x18slow_start_learning_rate\x18\n \x01(\t\x12\x14\n\x0cweight_decay\x18\x0b \x01(\t\x12 \n\x18learning_rate_decay_step\x18\x0c \x01(\t\x12\x17\n\x0fslow_start_step\x18\r \x01(\t\x12\x17\n\x0flearning_policy\x18\x0e \x01(\t\x12\x18\n\x10max_scale_factor\x18\x0f \x01(\t\x12\x18\n\x10min_scale_factor\x18\x10 \x01(\t\x12\x1e\n\x16scale_factor_step_size\x18\x11 \x01(\t\x12\x1d\n\x15\x64\x65\x63oder_output_stride\x18\x12 \x02(\x05\x12\x15\n\routput_stride\x18\x13 \x02(\x05\x12\x15\n\rmodel_variant\x18\x14 \x02(\t\x12\x14\n\x0c\x61trous_rates\x18\x15 \x03(\x05\x12\x1a\n\x0btrain_split\x18\x16 \x01(\t:\x05train\x12\x1e\n\neval_split\x18! \x01(\t:\nvalidation\x12\x17\n\x07\x64\x61taset\x18\x17 \x01(\t:\x06\x63ustom\x12\x18\n\x10train_batch_size\x18\x18 \x02(\x05\x12 \n\x18training_number_of_steps\x18\x19 \x02(\x05\x12\x17\n\x0ftrain_crop_size\x18\x1a \x03(\x05\x12\x16\n\x0e\x65val_crop_size\x18\" \x03(\x05\x12\x1d\n\x0f\x64l_custom_train\x18\x1b \x01(\x05:\x04\x31\x33\x33\x33\x12\"\n\x14\x64l_custom_validation\x18\x1c \x01(\x05:\x04\x31\x33\x33\x33\x12#\n\x15save_summaries_images\x18\x1d \x01(\x08:\x04true\x12\x1f\n\x12save_interval_secs\x18\x1e \x01(\x05:\x03\x36\x30\x30\x12\x1f\n\x13save_summaries_secs\x18\x1f \x01(\x05:\x02\x36\x30\x12\x15\n\nnum_clones\x18  \x01(\x05:\x01\x31')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_TRAININGPARAMETERS = _descriptor.Descriptor(
  name='TrainingParameters',
  full_name='deeplab.protos.TrainingParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='initialize_last_layer', full_name='deeplab.protos.TrainingParameters.initialize_last_layer', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='last_layers_contain_logits_only', full_name='deeplab.protos.TrainingParameters.last_layers_contain_logits_only', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='upsample_logits', full_name='deeplab.protos.TrainingParameters.upsample_logits', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='fine_tune_batch_norm', full_name='deeplab.protos.TrainingParameters.fine_tune_batch_norm', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='base_learning_rate', full_name='deeplab.protos.TrainingParameters.base_learning_rate', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='last_layer_gradient_multiplier', full_name='deeplab.protos.TrainingParameters.last_layer_gradient_multiplier', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='learning_power', full_name='deeplab.protos.TrainingParameters.learning_power', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='learning_rate_decay_factor', full_name='deeplab.protos.TrainingParameters.learning_rate_decay_factor', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='momentum', full_name='deeplab.protos.TrainingParameters.momentum', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='slow_start_learning_rate', full_name='deeplab.protos.TrainingParameters.slow_start_learning_rate', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='weight_decay', full_name='deeplab.protos.TrainingParameters.weight_decay', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='learning_rate_decay_step', full_name='deeplab.protos.TrainingParameters.learning_rate_decay_step', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='slow_start_step', full_name='deeplab.protos.TrainingParameters.slow_start_step', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='learning_policy', full_name='deeplab.protos.TrainingParameters.learning_policy', index=13,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='max_scale_factor', full_name='deeplab.protos.TrainingParameters.max_scale_factor', index=14,
      number=15, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='min_scale_factor', full_name='deeplab.protos.TrainingParameters.min_scale_factor', index=15,
      number=16, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='scale_factor_step_size', full_name='deeplab.protos.TrainingParameters.scale_factor_step_size', index=16,
      number=17, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='decoder_output_stride', full_name='deeplab.protos.TrainingParameters.decoder_output_stride', index=17,
      number=18, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='output_stride', full_name='deeplab.protos.TrainingParameters.output_stride', index=18,
      number=19, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='model_variant', full_name='deeplab.protos.TrainingParameters.model_variant', index=19,
      number=20, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='atrous_rates', full_name='deeplab.protos.TrainingParameters.atrous_rates', index=20,
      number=21, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='train_split', full_name='deeplab.protos.TrainingParameters.train_split', index=21,
      number=22, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("train").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='eval_split', full_name='deeplab.protos.TrainingParameters.eval_split', index=22,
      number=33, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("validation").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dataset', full_name='deeplab.protos.TrainingParameters.dataset', index=23,
      number=23, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("custom").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='train_batch_size', full_name='deeplab.protos.TrainingParameters.train_batch_size', index=24,
      number=24, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='training_number_of_steps', full_name='deeplab.protos.TrainingParameters.training_number_of_steps', index=25,
      number=25, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='train_crop_size', full_name='deeplab.protos.TrainingParameters.train_crop_size', index=26,
      number=26, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='eval_crop_size', full_name='deeplab.protos.TrainingParameters.eval_crop_size', index=27,
      number=34, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dl_custom_train', full_name='deeplab.protos.TrainingParameters.dl_custom_train', index=28,
      number=27, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1333,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dl_custom_validation', full_name='deeplab.protos.TrainingParameters.dl_custom_validation', index=29,
      number=28, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1333,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='save_summaries_images', full_name='deeplab.protos.TrainingParameters.save_summaries_images', index=30,
      number=29, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='save_interval_secs', full_name='deeplab.protos.TrainingParameters.save_interval_secs', index=31,
      number=30, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=600,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='save_summaries_secs', full_name='deeplab.protos.TrainingParameters.save_summaries_secs', index=32,
      number=31, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=60,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='num_clones', full_name='deeplab.protos.TrainingParameters.num_clones', index=33,
      number=32, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=60,
  serialized_end=1063,
)

DESCRIPTOR.message_types_by_name['TrainingParameters'] = _TRAININGPARAMETERS

TrainingParameters = _reflection.GeneratedProtocolMessageType('TrainingParameters', (_message.Message,), dict(
  DESCRIPTOR = _TRAININGPARAMETERS,
  __module__ = 'rastervision.protos.deeplab.train_pb2'
  # @@protoc_insertion_point(class_scope:deeplab.protos.TrainingParameters)
  ))
_sym_db.RegisterMessage(TrainingParameters)


# @@protoc_insertion_point(module_scope)



# flake8: noqa

from rastervision.filesystem.filesystem import (
    FileSystem, NotReadableError, NotWritableError, ProtobufParseException)
from rastervision.filesystem.local_filesystem import LocalFileSystem
from rastervision.filesystem.s3_filesystem import S3FileSystem
from rastervision.filesystem.http_filesystem import HttpFileSystem



import os
from copy import deepcopy

import rastervision as rv
from rastervision.core import ClassMap
from rastervision.evaluation \
    import (EvaluatorConfig, EvaluatorConfigBuilder)
from rastervision.protos.evaluator_pb2 import EvaluatorConfig as EvaluatorConfigMsg


class ClassificationEvaluatorConfig(EvaluatorConfig):
    """Abstract class for usage with simple evaluators that
    are classification-based.
    """

    def __init__(self,
                 evaluator_type,
                 class_map,
                 output_uri=None,
                 vector_output_uri=None):
        super().__init__(evaluator_type)
        self.class_map = class_map
        self.output_uri = output_uri
        self.vector_output_uri = vector_output_uri

    def to_proto(self):
        sub_msg = EvaluatorConfigMsg.ClassificationEvaluatorConfig(
            class_items=self.class_map.to_proto(),
            output_uri=self.output_uri,
            vector_output_uri=self.vector_output_uri)
        msg = EvaluatorConfigMsg(
            evaluator_type=self.evaluator_type, classification_config=sub_msg)

        return msg

    def update_for_command(self, command_type, experiment_config,
                           context=None):
        if command_type == rv.EVAL:
            if not self.output_uri:
                self.output_uri = os.path.join(experiment_config.eval_uri,
                                               'eval.json')
            if not self.vector_output_uri:
                self.vector_output_uri = os.path.join(
                    experiment_config.eval_uri, 'vector-eval.json')

    def report_io(self, command_type, io_def):
        if command_type == rv.EVAL:
            io_def.add_output(self.output_uri)
            io_def.add_output(self.vector_output_uri)

        return io_def


class ClassificationEvaluatorConfigBuilder(EvaluatorConfigBuilder):
    def __init__(self, cls, prev=None):
        self.config = {}
        if prev:
            self.config = {
                'output_uri': prev.output_uri,
                'vector_output_uri': prev.vector_output_uri,
                'class_map': prev.class_map
            }
        super().__init__(cls, self.config)

    def validate(self):
        if self.config.get('class_map') is None:
            raise rv.ConfigError(
                'class_map not set for ClassificationEvaluatorConfig. '
                'Use "with_class_map".')
        if not isinstance(self.config.get('class_map'), ClassMap):
            raise rv.ConfigError(
                'class_map set with "with_class_map" must be of type ClassMap, got {}'.
                format(type(self.config.get('class_map'))))

    @classmethod
    def from_proto(cls, msg):
        b = cls()
        return b.with_output_uri(msg.classification_config.output_uri) \
                .with_vector_output_uri(msg.classification_config.vector_output_uri) \
                .with_class_map(list(msg.classification_config.class_items))

    def with_output_uri(self, output_uri):
        """Set the output_uri.

            Args:
                output_uri: URI to the stats json to use
        """
        b = deepcopy(self)
        b.config['output_uri'] = output_uri
        return b

    def with_vector_output_uri(self, vector_output_uri):
        """Set the vector_output_uri.

            Args:
                vector_output_uri: URI to the vector stats json to use
        """
        b = deepcopy(self)
        b.config['vector_output_uri'] = vector_output_uri
        return b

    def with_task(self, task):
        """Sets a specific task type, e.g. rv.OBJECT_DETECTION."""
        if not hasattr(task, 'class_map'):
            raise rv.ConfigError('This evaluator requires a task '
                                 'that has a class_map property')
        return self.with_class_map(task.class_map)

    def with_class_map(self, class_map):
        """Set the class map to be used for evaluation.

            Args:
                class_map: The class map to be used

        """
        b = deepcopy(self)
        b.config['class_map'] = ClassMap.construct_from(class_map)
        return b



from abc import abstractmethod
import os
from copy import deepcopy

import rastervision as rv
from rastervision.core import (Config, ConfigBuilder, BundledConfigMixin)
from rastervision.protos.task_pb2 import TaskConfig as TaskConfigMsg


class TaskConfig(BundledConfigMixin, Config):
    def __init__(self,
                 task_type,
                 predict_batch_size=10,
                 predict_package_uri=None,
                 debug=True,
                 predict_debug_uri=None):
        self.task_type = task_type
        self.predict_batch_size = predict_batch_size
        self.predict_package_uri = predict_package_uri
        self.debug = debug
        self.predict_debug_uri = predict_debug_uri

    @abstractmethod
    def create_task(self, backend):
        """Create the Task that this configuration represents

           Args:
              backend: The backend to be used by the task.
        """
        pass

    def to_builder(self):
        return rv._registry.get_config_builder(rv.TASK, self.task_type)(self)

    def to_proto(self):
        return TaskConfigMsg(
            task_type=self.task_type,
            predict_batch_size=self.predict_batch_size,
            predict_package_uri=self.predict_package_uri,
            debug=self.debug,
            predict_debug_uri=self.predict_debug_uri)

    @staticmethod
    def builder(task_type):
        return rv._registry.get_config_builder(rv.TASK, task_type)()

    @staticmethod
    def from_proto(msg):
        """Creates a TaskConfig from the specificed protobuf message
        """
        return rv._registry.get_config_builder(rv.TASK, msg.task_type)() \
                           .from_proto(msg) \
                           .build()

    def update_for_command(self, command_type, experiment_config,
                           context=None):
        if command_type == rv.BUNDLE:
            if not self.predict_package_uri:
                self.predict_package_uri = os.path.join(
                    experiment_config.bundle_uri, 'predict_package.zip')

    def report_io(self, command_type, io_def):
        if command_type == rv.BUNDLE:
            io_def.add_output(self.predict_package_uri)


class TaskConfigBuilder(ConfigBuilder):
    def from_proto(self, msg):
        b = self.with_predict_batch_size(msg.predict_batch_size)
        b = b.with_predict_package_uri(msg.predict_package_uri)
        b = b.with_debug(msg.debug)
        b = b.with_predict_debug_uri(msg.predict_debug_uri)
        return b

    def with_predict_batch_size(self, predict_batch_size):
        """Sets the batch size to use during prediction."""
        b = deepcopy(self)
        b.config['predict_batch_size'] = predict_batch_size
        return b

    def with_predict_package_uri(self, predict_package_uri):
        """Sets the URI to save a predict package URI to during bundle."""
        b = deepcopy(self)
        b.config['predict_package_uri'] = predict_package_uri
        return b

    def with_debug(self, debug):
        """Flag for producing debug products."""
        b = deepcopy(self)
        b.config['debug'] = debug
        return b

    def with_predict_debug_uri(self, predict_debug_uri):
        """Set the directory to place prediction debug images"""
        b = deepcopy(self)
        b.config['predict_debug_uri'] = predict_debug_uri
        return b



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/tf_object_detection/graph_rewriter.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/tf_object_detection/graph_rewriter.proto',
  package='rastervision.protos.tf_object_detection',
  syntax='proto2',
  serialized_pb=_b('\n<rastervision/protos/tf_object_detection/graph_rewriter.proto\x12\'rastervision.protos.tf_object_detection\"\\\n\rGraphRewriter\x12K\n\x0cquantization\x18\x01 \x01(\x0b\x32\x35.rastervision.protos.tf_object_detection.Quantization\"Y\n\x0cQuantization\x12\x15\n\x05\x64\x65lay\x18\x01 \x01(\x05:\x06\x35\x30\x30\x30\x30\x30\x12\x16\n\x0bweight_bits\x18\x02 \x01(\x05:\x01\x38\x12\x1a\n\x0f\x61\x63tivation_bits\x18\x03 \x01(\x05:\x01\x38')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_GRAPHREWRITER = _descriptor.Descriptor(
  name='GraphRewriter',
  full_name='rastervision.protos.tf_object_detection.GraphRewriter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='quantization', full_name='rastervision.protos.tf_object_detection.GraphRewriter.quantization', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=105,
  serialized_end=197,
)


_QUANTIZATION = _descriptor.Descriptor(
  name='Quantization',
  full_name='rastervision.protos.tf_object_detection.Quantization',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='delay', full_name='rastervision.protos.tf_object_detection.Quantization.delay', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=500000,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='weight_bits', full_name='rastervision.protos.tf_object_detection.Quantization.weight_bits', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=8,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='activation_bits', full_name='rastervision.protos.tf_object_detection.Quantization.activation_bits', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=8,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=199,
  serialized_end=288,
)

_GRAPHREWRITER.fields_by_name['quantization'].message_type = _QUANTIZATION
DESCRIPTOR.message_types_by_name['GraphRewriter'] = _GRAPHREWRITER
DESCRIPTOR.message_types_by_name['Quantization'] = _QUANTIZATION

GraphRewriter = _reflection.GeneratedProtocolMessageType('GraphRewriter', (_message.Message,), dict(
  DESCRIPTOR = _GRAPHREWRITER,
  __module__ = 'rastervision.protos.tf_object_detection.graph_rewriter_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.GraphRewriter)
  ))
_sym_db.RegisterMessage(GraphRewriter)

Quantization = _reflection.GeneratedProtocolMessageType('Quantization', (_message.Message,), dict(
  DESCRIPTOR = _QUANTIZATION,
  __module__ = 'rastervision.protos.tf_object_detection.graph_rewriter_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.Quantization)
  ))
_sym_db.RegisterMessage(Quantization)


# @@protoc_insertion_point(module_scope)



from os.path import join
import tempfile

import pycocotools
from pycocotools.coco import COCO
from pycocotools.cocoeval import COCOeval
import numpy as np

from rastervision.utils.files import json_to_file


def get_coco_gt(targets, num_labels):
    images = []
    annotations = []
    ann_id = 1
    for img_id, target in enumerate(targets, 1):
        # Use fake height, width, and filename because they don't matter.
        images.append({
            'id': img_id,
            'height': 1000,
            'width': 1000,
            'file_name': '{}.png'.format(img_id)
        })
        boxes, labels = target.boxes, target.get_field('labels')
        for box, label in zip(boxes, labels):
            box = box.float().tolist()
            label = label.item()
            annotations.append({
                'id':
                ann_id,
                'image_id':
                img_id,
                'category_id':
                label,
                'area': (box[2] - box[0]) * (box[3] - box[1]),
                'bbox': [box[1], box[0], box[3] - box[1], box[2] - box[0]],
                'iscrowd':
                0
            })
            ann_id += 1

    categories = [{
        'id': label,
        'name': str(label),
        'supercategory': 'super'
    } for label in range(num_labels)]
    coco = {
        'images': images,
        'annotations': annotations,
        'categories': categories
    }
    return coco


def get_coco_preds(outputs):
    preds = []
    for img_id, output in enumerate(outputs, 1):
        for box, label, score in zip(output.boxes, output.get_field('labels'),
                                     output.get_field('scores')):
            box = box.float().tolist()
            label = label.item()
            score = score.item()
            preds.append({
                'image_id':
                img_id,
                'category_id':
                label,
                'bbox': [box[1], box[0], box[3] - box[1], box[2] - box[0]],
                'score':
                score
            })
    return preds


def compute_coco_eval(outputs, targets, num_labels):
    """Return mAP averaged over 0.5-0.95 using pycocotools eval.

    Note: boxes are in (ymin, xmin, ymax, xmax) format with values ranging
        from 0 to h or w.

    Args:
        outputs: (list) of length m containing dicts of form
            {'boxes': <tensor with shape (n, 4)>,
             'labels': <tensor with shape (n,)>,
             'scores': <tensor with shape (n,)>}
        targets: (list) of length m containing dicts of form
            {'boxes': <tensor with shape (n, 4)>,
             'labels': <tensor with shape (n,)>}
    """
    with tempfile.TemporaryDirectory() as tmp_dir:
        preds = get_coco_preds(outputs)
        # ap is undefined when there are no predicted boxes
        if len(preds) == 0:
            return None

        gt = get_coco_gt(targets, num_labels)
        gt_path = join(tmp_dir, 'gt.json')
        json_to_file(gt, gt_path)
        coco_gt = COCO(gt_path)

        pycocotools.coco.unicode = None
        coco_preds = coco_gt.loadRes(preds)

        ann_type = 'bbox'
        coco_eval = COCOeval(coco_gt, coco_preds, ann_type)

        coco_eval.evaluate()
        coco_eval.accumulate()
        coco_eval.summarize()

        return coco_eval


def compute_class_f1(coco_eval):
    precision = coco_eval.eval['precision'][0, :, :, 0, -1]
    scores = coco_eval.eval['scores'][0, :, :, 0, -1]
    recall = np.linspace(0, 1, num=precision.shape[0])
    recall = recall[:, None]

    f1s = (2 * precision * recall) / (np.maximum(precision + recall, 1e-4))
    best_f1s = f1s.max(axis=0)
    best_f1_inds = f1s.argmax(axis=0)
    best_scores = scores[best_f1_inds, range(len(best_f1_inds))]

    return best_f1s, best_scores



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/tf_object_detection/pipeline.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from rastervision.protos.tf_object_detection import eval_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_eval__pb2
from rastervision.protos.tf_object_detection import graph_rewriter_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_graph__rewriter__pb2
from rastervision.protos.tf_object_detection import input_reader_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_input__reader__pb2
from rastervision.protos.tf_object_detection import model_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_model__pb2
from rastervision.protos.tf_object_detection import train_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_train__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/tf_object_detection/pipeline.proto',
  package='rastervision.protos.tf_object_detection',
  syntax='proto2',
  serialized_pb=_b('\n6rastervision/protos/tf_object_detection/pipeline.proto\x12\'rastervision.protos.tf_object_detection\x1a\x32rastervision/protos/tf_object_detection/eval.proto\x1a<rastervision/protos/tf_object_detection/graph_rewriter.proto\x1a:rastervision/protos/tf_object_detection/input_reader.proto\x1a\x33rastervision/protos/tf_object_detection/model.proto\x1a\x33rastervision/protos/tf_object_detection/train.proto\"\xf5\x03\n\x17TrainEvalPipelineConfig\x12\x46\n\x05model\x18\x01 \x01(\x0b\x32\x37.rastervision.protos.tf_object_detection.DetectionModel\x12J\n\x0ctrain_config\x18\x02 \x01(\x0b\x32\x34.rastervision.protos.tf_object_detection.TrainConfig\x12P\n\x12train_input_reader\x18\x03 \x01(\x0b\x32\x34.rastervision.protos.tf_object_detection.InputReader\x12H\n\x0b\x65val_config\x18\x04 \x01(\x0b\x32\x33.rastervision.protos.tf_object_detection.EvalConfig\x12O\n\x11\x65val_input_reader\x18\x05 \x03(\x0b\x32\x34.rastervision.protos.tf_object_detection.InputReader\x12N\n\x0egraph_rewriter\x18\x06 \x01(\x0b\x32\x36.rastervision.protos.tf_object_detection.GraphRewriter*\t\x08\xe8\x07\x10\x80\x80\x80\x80\x02')
  ,
  dependencies=[rastervision_dot_protos_dot_tf__object__detection_dot_eval__pb2.DESCRIPTOR,rastervision_dot_protos_dot_tf__object__detection_dot_graph__rewriter__pb2.DESCRIPTOR,rastervision_dot_protos_dot_tf__object__detection_dot_input__reader__pb2.DESCRIPTOR,rastervision_dot_protos_dot_tf__object__detection_dot_model__pb2.DESCRIPTOR,rastervision_dot_protos_dot_tf__object__detection_dot_train__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_TRAINEVALPIPELINECONFIG = _descriptor.Descriptor(
  name='TrainEvalPipelineConfig',
  full_name='rastervision.protos.tf_object_detection.TrainEvalPipelineConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='model', full_name='rastervision.protos.tf_object_detection.TrainEvalPipelineConfig.model', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='train_config', full_name='rastervision.protos.tf_object_detection.TrainEvalPipelineConfig.train_config', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='train_input_reader', full_name='rastervision.protos.tf_object_detection.TrainEvalPipelineConfig.train_input_reader', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='eval_config', full_name='rastervision.protos.tf_object_detection.TrainEvalPipelineConfig.eval_config', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='eval_input_reader', full_name='rastervision.protos.tf_object_detection.TrainEvalPipelineConfig.eval_input_reader', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='graph_rewriter', full_name='rastervision.protos.tf_object_detection.TrainEvalPipelineConfig.graph_rewriter', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=True,
  syntax='proto2',
  extension_ranges=[(1000, 536870912), ],
  oneofs=[
  ],
  serialized_start=380,
  serialized_end=881,
)

_TRAINEVALPIPELINECONFIG.fields_by_name['model'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_model__pb2._DETECTIONMODEL
_TRAINEVALPIPELINECONFIG.fields_by_name['train_config'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_train__pb2._TRAINCONFIG
_TRAINEVALPIPELINECONFIG.fields_by_name['train_input_reader'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_input__reader__pb2._INPUTREADER
_TRAINEVALPIPELINECONFIG.fields_by_name['eval_config'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_eval__pb2._EVALCONFIG
_TRAINEVALPIPELINECONFIG.fields_by_name['eval_input_reader'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_input__reader__pb2._INPUTREADER
_TRAINEVALPIPELINECONFIG.fields_by_name['graph_rewriter'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_graph__rewriter__pb2._GRAPHREWRITER
DESCRIPTOR.message_types_by_name['TrainEvalPipelineConfig'] = _TRAINEVALPIPELINECONFIG

TrainEvalPipelineConfig = _reflection.GeneratedProtocolMessageType('TrainEvalPipelineConfig', (_message.Message,), dict(
  DESCRIPTOR = _TRAINEVALPIPELINECONFIG,
  __module__ = 'rastervision.protos.tf_object_detection.pipeline_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.TrainEvalPipelineConfig)
  ))
_sym_db.RegisterMessage(TrainEvalPipelineConfig)


# @@protoc_insertion_point(module_scope)



import io
import os
import subprocess
from datetime import datetime
from urllib.parse import urlparse

from rastervision.filesystem import (FileSystem, NotReadableError,
                                     NotWritableError)


# Code from https://alexwlchan.net/2017/07/listing-s3-keys/
def get_matching_s3_objects(bucket, prefix='', suffix='',
                            request_payer='None'):
    """
    Generate objects in an S3 bucket.

    :param bucket: Name of the S3 bucket.
    :param prefix: Only fetch objects whose key starts with
        this prefix (optional).
    :param suffix: Only fetch objects whose keys end with
        this suffix (optional).
    """
    import boto3
    s3 = boto3.client('s3')
    kwargs = {'Bucket': bucket, 'RequestPayer': request_payer}

    # If the prefix is a single string (not a tuple of strings), we can
    # do the filtering directly in the S3 API.
    if isinstance(prefix, str):
        kwargs['Prefix'] = prefix

    while True:

        # The S3 API response is a large blob of metadata.
        # 'Contents' contains information about the listed objects.
        resp = s3.list_objects_v2(**kwargs)

        try:
            contents = resp['Contents']
        except KeyError:
            return

        for obj in contents:
            key = obj['Key']
            if key.startswith(prefix) and key.endswith(suffix):
                yield obj

        # The S3 API is paginated, returning up to 1000 keys at a time.
        # Pass the continuation token into the next response, until we
        # reach the final page (when this field is missing).
        try:
            kwargs['ContinuationToken'] = resp['NextContinuationToken']
        except KeyError:
            break


def get_matching_s3_keys(bucket, prefix='', suffix='', request_payer='None'):
    """
    Generate the keys in an S3 bucket.

    :param bucket: Name of the S3 bucket.
    :param prefix: Only fetch keys that start with this prefix (optional).
    :param suffix: Only fetch keys that end with this suffix (optional).
    """
    for obj in get_matching_s3_objects(bucket, prefix, suffix, request_payer):
        yield obj['Key']


class S3FileSystem(FileSystem):
    @staticmethod
    def get_request_payer():
        # Import here to avoid circular reference.
        from rastervision.rv_config import RVConfig
        rv_config = RVConfig.get_instance()
        s3_config = rv_config.get_subconfig('AWS_S3')

        # 'None' needs the quotes because boto3 cannot handle None.
        return ('requester' if s3_config(
            'requester_pays', parser=bool, default='False') else 'None')

    @staticmethod
    def get_session():
        # Lazily load boto
        import boto3
        return boto3.Session()

    @staticmethod
    def matches_uri(uri: str, mode: str) -> bool:
        parsed_uri = urlparse(uri)
        return parsed_uri.scheme == 's3'

    @staticmethod
    def file_exists(uri: str, include_dir: bool = True) -> bool:
        # Lazily load boto
        import botocore

        parsed_uri = urlparse(uri)
        bucket = parsed_uri.netloc
        key = parsed_uri.path[1:]
        request_payer = S3FileSystem.get_request_payer()

        if include_dir:
            s3 = S3FileSystem.get_session().client('s3')
            try:
                # Ensure key ends in slash so that this won't pick up files that
                # contain the key as a prefix, but aren't actually directories.
                # Example: if key is 'model' then we don't want to consider
                # model-123 a match.
                dir_key = key if key[-1] == '/' else key + '/'
                response = s3.list_objects_v2(
                    Bucket=bucket,
                    Prefix=dir_key,
                    MaxKeys=1,
                    RequestPayer=request_payer)
                if response['KeyCount'] == 0:
                    return S3FileSystem.file_exists(uri, include_dir=False)
                return True
            except botocore.exceptions.ClientError as e:
                return False
        else:
            s3r = S3FileSystem.get_session().resource('s3')
            try:
                s3r.Object(bucket, key).load(RequestPayer=request_payer)
                return True
            except botocore.exceptions.ClientError as e:
                return False

    @staticmethod
    def read_str(uri: str) -> str:
        return S3FileSystem.read_bytes(uri).decode('utf-8')

    @staticmethod
    def read_bytes(uri: str) -> bytes:
        import botocore

        s3 = S3FileSystem.get_session().client('s3')
        request_payer = S3FileSystem.get_request_payer()

        parsed_uri = urlparse(uri)
        with io.BytesIO() as file_buffer:
            try:
                s3.download_fileobj(
                    parsed_uri.netloc,
                    parsed_uri.path[1:],
                    file_buffer,
                    ExtraArgs={'RequestPayer': request_payer})
                return file_buffer.getvalue()
            except botocore.exceptions.ClientError as e:
                raise NotReadableError('Could not read {}'.format(uri)) from e

    @staticmethod
    def write_str(uri: str, data: str) -> None:
        data = bytes(data, encoding='utf-8')
        S3FileSystem.write_bytes(uri, data)

    @staticmethod
    def write_bytes(uri: str, data: bytes) -> None:
        s3 = S3FileSystem.get_session().client('s3')

        parsed_uri = urlparse(uri)
        bucket = parsed_uri.netloc
        key = parsed_uri.path[1:]
        with io.BytesIO(data) as str_buffer:
            try:
                s3.upload_fileobj(str_buffer, bucket, key)
            except Exception as e:
                raise NotWritableError('Could not write {}'.format(uri)) from e

    @staticmethod
    def sync_from_dir(src_dir_uri: str,
                      dest_dir_uri: str,
                      delete: bool = False) -> None:  # pragma: no cover
        command = ['aws', 's3', 'sync', src_dir_uri, dest_dir_uri]
        if delete:
            command.append('--delete')
        subprocess.run(command)

    @staticmethod
    def sync_to_dir(src_dir_uri: str, dest_dir_uri: str,
                    delete: bool = False) -> None:  # pragma: no cover
        command = ['aws', 's3', 'sync', src_dir_uri, dest_dir_uri]
        if delete:
            command.append('--delete')
        request_payer = S3FileSystem.get_request_payer()
        if request_payer:
            command.append('--request-payer')
        subprocess.run(command)

    @staticmethod
    def copy_to(src_path: str, dst_uri: str) -> None:
        s3 = S3FileSystem.get_session().client('s3')

        parsed_uri = urlparse(dst_uri)
        if os.path.isfile(src_path):
            try:
                s3.upload_file(src_path, parsed_uri.netloc,
                               parsed_uri.path[1:])
            except Exception as e:
                raise NotWritableError(
                    'Could not write {}'.format(dst_uri)) from e
        else:
            S3FileSystem.sync_to_dir(src_path, dst_uri, delete=True)

    @staticmethod
    def copy_from(uri: str, path: str) -> None:
        import botocore

        s3 = S3FileSystem.get_session().client('s3')
        request_payer = S3FileSystem.get_request_payer()

        parsed_uri = urlparse(uri)
        try:
            s3.download_file(
                parsed_uri.netloc,
                parsed_uri.path[1:],
                path,
                ExtraArgs={'RequestPayer': request_payer})
        except botocore.exceptions.ClientError:
            raise NotReadableError('Could not read {}'.format(uri))

    @staticmethod
    def local_path(uri: str, download_dir: str) -> None:
        parsed_uri = urlparse(uri)
        path = os.path.join(download_dir, 's3', parsed_uri.netloc,
                            parsed_uri.path[1:])
        return path

    @staticmethod
    def last_modified(uri: str) -> datetime:
        parsed_uri = urlparse(uri)
        bucket, key = parsed_uri.netloc, parsed_uri.path[1:]
        s3 = S3FileSystem.get_session().client('s3')
        request_payer = S3FileSystem.get_request_payer()
        head_data = s3.head_object(
            Bucket=bucket, Key=key, RequestPayer=request_payer)
        return head_data['LastModified']

    @staticmethod
    def list_paths(uri, ext=''):
        request_payer = S3FileSystem.get_request_payer()
        parsed_uri = urlparse(uri)
        bucket = parsed_uri.netloc
        prefix = os.path.join(parsed_uri.path[1:])
        keys = get_matching_s3_keys(
            bucket, prefix, suffix=ext, request_payer=request_payer)
        return [os.path.join('s3://', bucket, key) for key in keys]



import click

from rastervision.command import Command


class ChipCommand(Command):
    def __init__(self, command_config):
        self.command_config = command_config

    def run(self, tmp_dir=None):
        if not tmp_dir:
            tmp_dir = self.get_tmp_dir()
        msg = 'Making training chips...'
        click.echo(click.style(msg, fg='green'))

        cc = self.command_config

        backend = cc.backend.create_backend(cc.task)
        task = cc.task.create_task(backend)

        train_scenes = list(
            map(lambda s: s.create_scene(cc.task, tmp_dir), cc.train_scenes))

        val_scenes = list(
            map(lambda s: s.create_scene(cc.task, tmp_dir), cc.val_scenes))

        augmentors = list(map(lambda a: a.create_augmentor(), cc.augmentors))

        task.make_chips(train_scenes, val_scenes, augmentors, tmp_dir)



# flake8: noqa

from rastervision.evaluation.evaluation_item import *
from rastervision.evaluation.class_evaluation_item import *
from rastervision.evaluation.evaluator import *
from rastervision.evaluation.evaluator_config import *
from rastervision.evaluation.classification_evaluation import *
from rastervision.evaluation.chip_classification_evaluation import *
from rastervision.evaluation.object_detection_evaluation import *
from rastervision.evaluation.semantic_segmentation_evaluation import *
from rastervision.evaluation.classification_evaluator import *
from rastervision.evaluation.classification_evaluator_config import *
from rastervision.evaluation.chip_classification_evaluator import *
from rastervision.evaluation.chip_classification_evaluator_config import *
from rastervision.evaluation.object_detection_evaluator import *
from rastervision.evaluation.object_detection_evaluator_config import *
from rastervision.evaluation.semantic_segmentation_evaluator import *
from rastervision.evaluation.semantic_segmentation_evaluator_config import *



from abc import (ABC, abstractmethod)

import rastervision as rv


class EvaluatorDefaultProvider(ABC):
    @staticmethod
    @abstractmethod
    def is_default_for(task_type):
        """Returns True if this evaluator is the default for this tasks_type"""
        pass

    @abstractmethod
    def construct(task):
        """Constructs the default evaluator.
        """
        pass


class ObjectDetectionEvaluatorDefaultProvider(EvaluatorDefaultProvider):
    @staticmethod
    def is_default_for(task_type):
        return task_type == rv.OBJECT_DETECTION

    @staticmethod
    def construct(task):
        return rv.EvaluatorConfig.builder(rv.OBJECT_DETECTION_EVALUATOR) \
                                 .with_task(task) \
                                 .build()


class ChipClassificationEvaluatorDefaultProvider(EvaluatorDefaultProvider):
    @staticmethod
    def is_default_for(task_type):
        return task_type == rv.CHIP_CLASSIFICATION

    @staticmethod
    def construct(task):
        return rv.EvaluatorConfig.builder(rv.CHIP_CLASSIFICATION_EVALUATOR) \
                                 .with_task(task) \
                                 .build()


class SemanticSegmentationEvaluatorDefaultProvider(EvaluatorDefaultProvider):
    @staticmethod
    def is_default_for(task_type):
        return task_type == rv.SEMANTIC_SEGMENTATION

    @staticmethod
    def construct(task):
        return rv.EvaluatorConfig.builder(rv.SEMANTIC_SEGMENTATION_EVALUATOR) \
                                 .with_task(task) \
                                 .build()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/tf_object_detection/matcher.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from rastervision.protos.tf_object_detection import argmax_matcher_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_argmax__matcher__pb2
from rastervision.protos.tf_object_detection import bipartite_matcher_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_bipartite__matcher__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/tf_object_detection/matcher.proto',
  package='rastervision.protos.tf_object_detection',
  syntax='proto2',
  serialized_pb=_b('\n5rastervision/protos/tf_object_detection/matcher.proto\x12\'rastervision.protos.tf_object_detection\x1a<rastervision/protos/tf_object_detection/argmax_matcher.proto\x1a?rastervision/protos/tf_object_detection/bipartite_matcher.proto\"\xc4\x01\n\x07Matcher\x12P\n\x0e\x61rgmax_matcher\x18\x01 \x01(\x0b\x32\x36.rastervision.protos.tf_object_detection.ArgMaxMatcherH\x00\x12V\n\x11\x62ipartite_matcher\x18\x02 \x01(\x0b\x32\x39.rastervision.protos.tf_object_detection.BipartiteMatcherH\x00\x42\x0f\n\rmatcher_oneof')
  ,
  dependencies=[rastervision_dot_protos_dot_tf__object__detection_dot_argmax__matcher__pb2.DESCRIPTOR,rastervision_dot_protos_dot_tf__object__detection_dot_bipartite__matcher__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_MATCHER = _descriptor.Descriptor(
  name='Matcher',
  full_name='rastervision.protos.tf_object_detection.Matcher',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='argmax_matcher', full_name='rastervision.protos.tf_object_detection.Matcher.argmax_matcher', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='bipartite_matcher', full_name='rastervision.protos.tf_object_detection.Matcher.bipartite_matcher', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='matcher_oneof', full_name='rastervision.protos.tf_object_detection.Matcher.matcher_oneof',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=226,
  serialized_end=422,
)

_MATCHER.fields_by_name['argmax_matcher'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_argmax__matcher__pb2._ARGMAXMATCHER
_MATCHER.fields_by_name['bipartite_matcher'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_bipartite__matcher__pb2._BIPARTITEMATCHER
_MATCHER.oneofs_by_name['matcher_oneof'].fields.append(
  _MATCHER.fields_by_name['argmax_matcher'])
_MATCHER.fields_by_name['argmax_matcher'].containing_oneof = _MATCHER.oneofs_by_name['matcher_oneof']
_MATCHER.oneofs_by_name['matcher_oneof'].fields.append(
  _MATCHER.fields_by_name['bipartite_matcher'])
_MATCHER.fields_by_name['bipartite_matcher'].containing_oneof = _MATCHER.oneofs_by_name['matcher_oneof']
DESCRIPTOR.message_types_by_name['Matcher'] = _MATCHER

Matcher = _reflection.GeneratedProtocolMessageType('Matcher', (_message.Message,), dict(
  DESCRIPTOR = _MATCHER,
  __module__ = 'rastervision.protos.tf_object_detection.matcher_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.Matcher)
  ))
_sym_db.RegisterMessage(Matcher)


# @@protoc_insertion_point(module_scope)



import rastervision as rv
from rastervision.task import TaskConfig
from rastervision.backend import BackendConfig
from rastervision.analyzer import AnalyzerConfig
from rastervision.data import SceneConfig


def check_analyzers_type(analyzers):
    if not isinstance(analyzers, list):
        raise rv.ConfigError(
            'analyzers must be a list of AnalyzerConfig objects, got {}'.
            format(type(analyzers)))
    for analyzer in analyzers:
        if not issubclass(type(analyzer), AnalyzerConfig):
            if not isinstance(analyzer, str):
                raise rv.ConfigError(
                    'analyzers must be of class AnalyzerConfig or string, got {}'.
                    format(type(analyzer)))


def check_backend_type(backend):
    if not issubclass(type(backend), BackendConfig):
        raise rv.ConfigError(
            'Backend must be a child of class BackendConfig, got {}'.format(
                type(backend)))


def check_scenes_type(scenes):
    if not isinstance(scenes, list):
        raise rv.ConfigError(
            'scenes must be a list of SceneConfig objects, got {}'.format(
                type(scenes)))
    for scene in scenes:
        if not isinstance(scene, SceneConfig):
            if not isinstance(scene, str):
                raise rv.ConfigError(
                    'scene must be a SceneConfig object or str, got {}'.format(
                        type(scene)))


def check_task_type(task):
    if not issubclass(type(task), TaskConfig):
        raise rv.ConfigError(
            'Task must be a child class of TaskConfig, got {}'.format(
                type(task)))
